##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
	Property regression models

	Ability to use DMF csv files to regress important parameters 
	for literature based models and ALAMO models.

	Tracking property metadata and new model metadata
	* See :meth:'PropertyData.from_csv', for expected format for data.
	* See :meth:'PropertyMetadata()' for the expected format for metadata.

	Author: Marissa Engle
"""
##### WARNING WARNING WARNING WARNING WARNING WARNING
##### For now this module is going to get commented out.
##### It has not been tested in too long
##### WARNING WARNING WARNING WARNING WARNING WARNING

# # Local Imports
# from __future__ import division
# from pyomo.environ import *
# from pyomo.opt import SolverFactory
# # import idaes_dmf.propdb.types as prop_types # older version
#
# from idaes_dmf import dmf, resource, propdata
# from idaes_dmf.resource import Resource, FilePath;
# import pendulum
#
# import itertools
# import visPlot
# from idaes.surrogate import alamopy
#
# # Third-party imports
# try:
# 	import pandas as pd
# 	import numpy as np
# 	import sympy
# 	from sympy.parsing.sympy_parser import parse_expr
# except ImportError:
# 	np, pd = None, None
#
# # Change to your location of mydata in DMF
# my_dmf = dmf.DMF("/home/mengle/DMF/mydata/")
#
# class PropertyPackage(object):
# 	"Property package containing multiple Property Models"
#
# 	def __init__(self):
# 		self.textFile = None;
# 		self.build_package();
#
# 	def build_package(self):
# 		self.openFile();
# 		self.importData();
# 		self.buildModels();
# 		self.closeFile();
#
# 	# TODO: Write the property package resource
# 	# def create_dmf_PropertyPackageResource(self):
# 	# 	print 'TODO: Property Package'
#
# 	def openFile(self):
# 		print('Creating new text file');
# 		try:
# 			self.textFile = open('propTrial.py', 'w')
# 		except:
# 			print("couldn\'t open the file. Something wrong")
# 			sys.exit();
#
# 	def closeFile(self):
# 		""" Closes the GAMS file"""
# 		self.textFile.close();
# 		print('Closing the File');
#
# class PropertyModel(object):
#
# 	def __init__(self, data=None, metadata=None, outputFile = None, almFile = None, desc = None, **kwargs):
# 		self.textfile = outputFile;
# 		self.opt = SolverFactory('ipopt')
# 		self.model = ConcreteModel();
#
# 		self.alamo_model = ALAMOModel(data, metadata, almFile)
# 		self.alamo_model.set_name(almFile);
# 		self.table = propdata.PropertyTable(data=data, metadata=metadata)
# 		self.data = self.table.data
#
# 		#adds data resource
# 		# TODO: add if it doesn't exist
# 		self.create_dmf_DataResource(data, metadata,desc);
# 		# d.add(rsrc)
#
# 		names = data.names()
# 		# print list(names)
#
# 		# print self.alamo_model.xdata;
# 		# self.solve_model(self.model)
# 		if 'almopt' in kwargs:
# 			self.build_alamo_model(self.alamo_model, kwargs)
# 		else:
# 			self.build_alamo_model(self.alamo_model)
# 		# self.create_dmf_PropertyModelResoure();
#
# 		# result = self.choose(self.model, self.alamo_model)
#
# 	def create_dmf_DataResource(self, data=None, metadata= None, desc= None):
# 		self.table = propdata.PropertyTable(data=data, metadata=metadata)
# 		rsrc = resource.PropertyDataResource(self.table);
# 		spec = {'desc!':desc}
# 		if desc is not None:
# 			rsrc.desc = desc
#
# 		resources = my_dmf.find(spec)
# 		try:
# 			findResc = next(resources);
# 			print 'Existed Data Resource', desc
# 		except Exception, e:
# 			print "No Resource found end of generator issue"
# 		finally:
# 			print 'New Data Resource', desc
# 			my_dmf.add(rsrc)
# 		# my_dmf.remove(filter_dict={'desc':'test'})
# 		self.dataResource = rsrc
#
#
# 	def create_dmf_PropertyModelResoure(self):
# 		print 'Property Model Resource'
#
# 		rsrc = resource.Resource(type='property_model')
# 		rsrc.created = pendulum.now()
# 		rsrc.desc = 'property model for something'
# 		# rsrc.sources = [self.dataResource.sources];
# 		# rsrc.codes  = [resource.Code(type='python function', desc='ALAMO generated property model', name='SteamEnthalpyalm.py'),
# 				# resource.Code(type='python function', desc ='ALAMO generated parameter property model', name='SteamEnthalpycv.py')]
# 		# rsrc.datafiles = [self.dataResource.file];
# 		my_dmf.add(rsrc)
#
# 	def get_dataframe(self):
# 		df = self.data.values_dataframe();
# 		return df;
#
# 	def get_residuals(self):
# 		residuals = None;
# 		return residuals;
#
# 	def plot_residuals(self):
# 		# def plot_residual(flowsheet, idaes_model_object_name, x_name,
#                   # y_name, plot_title='Some default title',
#                   # save_directory='', file_format='png'):
# 		names = data.names();
#
# 		residuals = self.model.get_residuals();
# 		ALAMO_residuals = self.alamo_model.get_residuals();
#
# 		visPlot.plot_residual(None, self, 'X_label', 'residuals', "Residuals", "ResidualPlot");
#
# 	def build_model(self, model):
# 		self.data_sets()
# 		self.sets(model)
# 		self.parameters(model)
# 		self.variables(model)
# 		self.objective(model)
#
# 	def solve_model(self, model):
# 		self.build_model(model)
# 		self.results = self.opt.solve(model);
# 		# self.model_output(model, self.textfile);
#
# 		SSE_params = model.OBJ();
# 		self.results_str(model);
#
# 	def build_alamo_model(self, alamo_model):
# 		self.alamo_datasets(alamo_model);
# 		# self.alamo_datasets(alamo_model);
#
# 	def choose(self, model, alamo_model):
# 		# output for DMF and output of python file, need to change 1/10/2018
# 		AModels = alamo_model.models;
# 		results = min(AModels, key=lambda p: float(p['ssr']))
#
# 		SSE_params = model.OBJ();
# 		self.results_str(model);
#
# 		if SSE_params < float(results['ssr']):
# 			print "Write param model"
# 			self.model_output(model, self.textfile);
# 		else:
# 			self.alamo_output(results, self.textfile)
#
# class ALAMOModel(object):
#
# 	def __init__ (self,data, metadata, almname):
# 		self.xdata = []
# 		self.zdata = []
# 		self.models = [];
# 		self.almname = 'Property.alm'
#
#
# 	def get_residuals():
# 		residuals = None;
# 		return residuals;
#
# 	def set_name(self,almname):
# 		self.almname=almname;
#
# 	def build_alamo_model(self, xdata, zdata, xlabels = None, zlabels='e', **kwargs):
# 		# results=alamopy.doalamo(xdata,zdata,xlabels=xlbls,  zlabels='expr',  monomialpower='1,2,3', multi3power='1,2,3', multi2power='1,2,3', ratiopower='1 2', logfcns=1, expfcns=1, savescratch=1, expandoutput=1); #logfcns=1, expfcns=1, ratiopower='1 2 ', multi3power='1 2', multi2power='1 2 '
# 		# results=alamopy.doalamo(xdata,zdata,xlabels=xlbls,  zlabels='e',  monomialpower='1 2 3', multi3power='1 2 3', multi2power='1 2 3', savescratch=1, expandoutput=1); #logfcns=1, expfcns=1, ratiopower='1 2 ', multi3power='1 2', multi2power='1 2 '
# 		# results=alamopy.doalamo(xdata,zdata, almname='SteamEnthalpy.alm', almopt='steamcustom.alm', monomialpower='1 2 3', multi3power='1 2 3', multi2power='1 2 3', savescratch=1, expandoutput=1); #logfcns=1, expfcns=1, ratiopower='1 2 ', multi3power='1 2', multi2power='1 2 '
#
# 		if 'almopt' in kwargs:
# 			results=alamopy.doalamo(xdata,zdata, xlabels=xlabels,  zlabels='e', almname=self.almname, almopt=kwargs.get("almopt"), monomialpower='1 2 3', multi3power='1 2 3', multi2power='1 2 3', savescratch=1, expandoutput=1); #logfcns=1, expfcns=1, ratiopower='1 2 ', multi3power='1 2', multi2power='1 2 '
# 		else:
# 			results=alamopy.doalamo(xdata,zdata, almname=self.almname, monomialpower='1 2 3', multi3power='1 2 3', multi2power='1 2 3', savescratch=1, expandoutput=1); #logfcns=1, expfcns=1, ratiopower='1 2 ', multi3power='1 2', multi2power='1 2 '
#
# 		print "Model function: %s" % results['model'];
# 		print "Sum of squared error = %s" % results['ssr']
# 		print "R^2 = %s" % results['R2'];
# 		res = alamopy.almconfidence(results, xdata, zdata);
# 		print "Confidence Interval =%s" % results['conf_inv']
# 		alamopy.almplot(res)
# 		print results['model'].replace("^","**")
# 		alm_model = results['model']
# 		self.models.append(results);
#
# 	def enumerate_datasets(self, xvars, xlabels, xdata, **kwargs):
# 		xsets= range(0,xvars)
#
# 		xlabels = np.array(xlabels)
#
# 		# for i in range(1, xvars+1):
# 		# 	for subset in itertools.combinations(xsets, i):
# 		# 		sub = np.array(subset)
# 		# 		#print self.xdata[:,sub]
# 		# 		self.build_alamo_model(self.xdata[:,sub], self.zdata, tuple(xlabels[sub]),'expr')
# 		self.build_alamo_model(self.xdata[:,:], self.zdata, xlabels= xlabels, zlabels='e', **kwargs)
# 		#xdata.append(self.data.get_column('r').values)
#
# 	# def model_output(self,  model, textfile):
# 	# 	print "Write ALAMO Model"
# 	# 	textfile.write("def _V_liq(self): \n");
# 	# 	textfile.write("\t %s)\n\n" %model['model'])
#
#
# # Example with SpiraxSteamTableEnthalpy.csv
# class SteamPropertyPackage(PropertyPackage):
# 	def __init__(self):
# 		self.textFile = None;
# 		self.build_package();
#
# 	def importData(self):
# 		# self.data1 = propdata.PropertyData.from_csv("/home/mengle/DMF/mydata/SpiraxSteamTableEnthalpy.csv",2);
# 		# self.metadata1 = propdata.PropertyMetadata.from_csv("/home/mengle/DMF/mydata/SpiraxSteamTableEnthalpy-meta.csv")
#
# 		self.data1 = propdata.PropertyData.from_csv("/home/mengle/DMF/mydata/SpiraxEnthalpy.csv",2);
# 		self.metadata1 = propdata.PropertyMetadata.from_csv("/home/mengle/DMF/mydata/SpiraxEnthalpy-meta.csv")
#
# 	def buildModels(self):
# 		print "Enthalpy Model"
# 		Enthalpy = EnthalpyModel(self.data1, self.metadata1, self.textFile, "Enthalpy", "spirax enthalpy");
#
# class EnthalpyModel(PropertyModel):
#
#     def __init__(self, data, metadata, outputFile, almtrace, desc):
#     	PropertyModel.__init__(self,data, metadata, outputFile, almtrace, desc);
#
#     def results_str(self, model):
#     	print "Sum of squared error = %f" % model.OBJ();
#
#     def model_output(self, model, textfile):
#     	textfile.write("def _h_vap(self): \n")
#     	textfile.write("\tTref=PhysPropParams.Tref")
#     	textfile.write("\tPref=PhysPropParams.Pref")
#
#     def data_sets(self):
#     	self.I = np.asarray(range(self.data.num_rows))
#     	self.Tr = dict(zip(self.I,self.data.get_column('Tr').values))
#     	self.Pr = dict(zip(self.I,self.data.get_column('Pr').values))
#     	# self.h_vap = dict(zip(self.I,self.data.get_column('Jkg').values))
#     	self.h_vap = dict(zip(self.I,self.data.get_column('Jmol').values))
#
#
#     def sets(self,model):
#     	model.I = Set(initialize=self.I)
#
#     def parameters(self,model):
#     	# Given parameters of the system
#     	model.Tr = Param(model.I, initialize = self.Tr)
#     	model.Pr = Param(model.I, initialize = self.Pr)
#     	model.h_vap = Param(model.I, intialize  = self.h_vap)
#
#     def alamo_datasets(self,alamo_model):
#     	# zdata = np.asarray(self.data.get_column('Jkg').values)
#     	zdata = np.asarray(self.data.get_column('Jmol').values)
#     	self.alamo_model.zdata = zdata.reshape(len(zdata),1);
#
#     	xdata = []
#     	xdata.append(self.data.get_column('Tr').values);
#     	xdata.append(self.data.get_column('Pr').values);
#
#     	self.alamo_model.xdata = np.transpose(xdata);
#     	self.alamo_model.enumerate_datasets(2, ['Tr', 'Pr'], self.alamo_model.xdata)#, almopt='steamcustom.alm')
#
#     def alamo_output(self,alamo_model, textfile):
#     	textfile.write("def h_vap(self): \n")
#     	textfile.write("\t# ALAMO Model")
#
#
#
# 	# def alamo_output(self, alamo_model,textfile):
# 	# 	textfile.write("def _h_vap(self): \n");
# 	# 	textfile.write("\t# ALAMO Model\n")
# 	# 	textfile.write("\t(T, P, y) = self.deref_tpy()\n");
# 	# 	textfile.write("\tself.var_or_expr(name=\"visc_liq\", doc=\"Liquid molar volume\", expr=%s)\n\n" %alamo_model['model'])
#
# 	#     # def _h_vap(self):
# 	#     #     c = CpIdealGasParams.c
# 	#     #     Tref = PhysPropParams.Tref
# 	#     #     hv_h2o = 43594
# 	#     #     hv_mea = 58000
# 	#     #     (T, P, y) = self.deref_tpy()
# 	#     #     expr_h_vap = sum((
# 	#     #         (c[i][0]*T + c[i][1]/2*T**2 + c[i][2]/3*T*3 + c[i][3]/4*T**4) -
# 	#     #         (c[i][0]*Tref + c[i][1]/2*Tref**2 + c[i][2]/3*Tref*3 + c[i][3]/4*Tref**4)
# 	#     #         )*y[i] for i in y.keys()) + y["H2O"]*hv_h2o + y["MEA"]*hv_mea
# 	#     #     self.var_or_expr(name="h_vap", expr=expr_h_vap, doc="Ideal gas enthalpy")
#
#
#
# class MEAPropertyPackage(PropertyPackage):
#
# 	def __init__(self):
# 		self.textFile = None;
# 		self.build_package();
#
# 	def importData(self):
# 		d = dmf.DMF("/home/mengle/DMF/mydata/");
# 		self.data1 = propdata.PropertyData.from_csv('/home/mengle/DMF/examples/jayarathna-data.csv', 2)
# 		self.metadata1 = propdata.PropertyMetadata.from_csv('/home/mengle/DMF/examples/jayarathna-meta.csv')#, 'MEA')
#
# 		self.data2 = propdata.PropertyData.from_csv("/home/mengle/DMF/mydata/amundsen-data.csv", 2)
# 		self.metadata2 = propdata.PropertyMetadata.from_csv("/home/mengle/DMF/mydata/amundsen-meta.csv")
#
# 		self.data3 = propdata.PropertyData.from_csv("/home/mengle/DMF/mydata/han-data.csv", 2)
# 		num_added = self.data3.add_csv("/home/mengle/DMF/mydata/jayarathna-data.csv")
# 		num_added = self.data3.add_csv("/home/mengle/DMF/mydata/amundsen-data2.csv")
# 		self.metadata3 = propdata.PropertyMetadata.from_csv("/home/mengle/DMF/mydata/han-meta.csv")
#
# 	def buildModels(self):
# 		print "Surface Tension Model"
# 		SurfaceTension = SurfaceTensionModel(self.data1, self.metadata1, self.textFile, "SurfaceTension", "jayrathna surface tension");
# 		# SurfaceTension.get_dataframe();
#
# 		print "Viscosity Model"
# 		Viscosity = ViscosityModel(self.data2,self.metadata2, self.textFile, "Viscosity", "amundsen viscosity")
#
# 		print "Density Model"
# 		Density = DensityModel(self.data3, self.metadata3, self.textFile, "Density", "amundsen jayrathna hans density")
#
# class DensityModel(PropertyModel):
#
# 	# class VLiqParams(object):
# 		#    # Liquid molar volume parameters (Weiland et al. 1998, adjusted for si units)
# 		#    a = {'MEA':-5.35162e-4, 'H2O':-3.2484e-3, 'CO2':0}
# 		#    b = {'MEA':-0.451417, 'H2O':1.65, 'CO2':0}
# 		#    c = {'MEA':1194.51, 'H2O':793.0, 'CO2':927.0}
# 		#    Vst = -1.8218e-6
#
# 		# def _V_liq(self):
# 	    #    (T, P, y) = self.deref_tpy()
# 	    #    a = VLiqParams.a
# 	    #    b = VLiqParams.b
# 	    #    c = VLiqParams.c
# 	    #    Vst = VLiqParams.Vst
# 	    #    mw = PhysPropParams.mw
# 	    #    if self.external:
# 	    #        self.var_or_expr(name="V_liq2",
# 	    #            expr=self.extf_V_liq(T, y['CO2'], y['H2O'], 0.0, y['MEA']),
# 	    #            doc="Liquid molar volume")
# 	    #        return
# 	    #    self.var_or_expr(name="V_liq", doc="Liquid molar volume",
# 	    #        expr=sum(y[i]*mw[i]/(a[i]*T**2 + b[i]*T + c[i]) for i in y.keys()) +
# 	    #        y['MEA']*y['H2O']*Vst)
#
# 	def __init__(self, data, metadata, outputFile, almtrace, description):
# 		PropertyModel.__init__(self, data, metadata, outputFile, almtrace, description);
#
#
# 	def results_str(self, model):
# 		print model.a_mea.value, model.b_mea.value, model.c_mea.value, model.c_CO2.value, model.Vstar.value;
# 		print "Sum of squared error = %f" % model.OBJ();
#
#
# 	def model_output(self, model, textfile):
# 		textfile.write("def _V_liq(self): \n");
# 		textfile.write("\t(T, P, y) = self.deref_tpy()\n");
# 		for v in model.component_objects(Var, active = True):
# 			textfile.write("\t%s = %s\n" %(v.name, v.value))
# 		textfile.write("\tVst = -1.8218e-6\n")
# 		textfile.write("\tmw = PhysPropParams.mw\n")
# 		textfile.write("\tr = self.wt_frac[\"MEA\"]\n\n");
# 		expr = "sum(y[i]*mw[i]/(a[i]*T**2 + b[i]*T + c[i]) for i in y.keys()) + y['MEA']*y['H2O']*Vst)"
# 		textfile.write("\tself.var_or_expr(name=\"V_liq\", doc=\"Liquid molar volume\", expr=%s)\n\n" %expr)
#
# 	def data_sets(self):
# 		self.I = np.asarray(range(self.data.num_rows));
# 		self.Temp = dict(zip(self.I,self.data.get_column('T').values));
# 		self.alpha = dict(zip(self.I,self.data.get_column('CO2 Loading').values));
# 		self.rstar = dict(zip(self.I,self.data.get_column('r').values));
# 		self.rho = dict(zip(self.I,self.data.get_column('Density Data').values));
#
# 	def sets(self, model):
# 		model.I = Set(initialize=self.I)
#
# 	def parameters(self, model):
# 		# Given parameters of the system
# 		model.alpha = Param(model.I, initialize = self.alpha)
# 		model.r = Param(model.I, initialize=self.rstar)
# 		model.T = Param(model.I, initialize=self.Temp)
#
# 		# Measured values of interest
# 		model.rho_sln = Param(model.I, initialize=self.rho)
#
# 	def variables(self, model):
# 		# Regressed parameter values
# 		model.a_mea = Var(domain=Reals, initialize = 1)
# 		model.b_mea = Var(domain= Reals, initialize = 1)
# 		model.c_mea = Var(domain=Reals, initialize = 1)
#
# 		#model.a_CO2 = Var(domain=Reals, initialize = 0)
# 		#model.b_CO2 = Var(domain= Reals, initialize = 0)
# 		model.c_CO2 = Var(domain=Reals, initialize = 1)
#
# 		#model.d = Var(domain=Reals, initialize = 1)
# 		#model.e = Var(domain=Reals, initialize = 1)
# 		#model.V_CO2 = Var(domain=NonNegativeReals, initialize = 1)
# 		model.Vstar = Var(domain=Reals, initialize = -1)
#
# 	def objective(self, model):
#
# 		def obj_expression(model):
# 			expr = 0
#
# 			# Molecular Weight
# 			MW_mea = 61.08
# 			MW_H2O = 18.02
# 			MW_CO2 = 44.01
#
# 			for i in model.I:
#
# 				# Mol Fractions
# 				X_mea = 1/(1+ model.alpha[i] + (MW_mea/MW_H2O)*((1-model.r[i])/model.r[i]))
# 				X_CO2 = X_mea * model.alpha[i]
# 				X_H2O =  1 - X_mea - X_CO2
# 				MW_sln = X_mea*MW_mea + X_CO2*MW_CO2 + X_H2O*MW_H2O
#
# 				# Volume
# 				V_mea = MW_mea/((model.a_mea/1e7)*model.T[i]**2 + (model.b_mea/1e4)*model.T[i] + model.c_mea)
# 				V_CO2 = MW_CO2/model.c_CO2;
# 				#V_CO2 = MW_CO2/((model.a_CO2/1e7)*model.T[i]**2 + (model.b_CO2/1e4)*model.T[i] + model.c_CO2)
# 				V_H2O = MW_H2O/(-(32.484/1e7)*model.T[i]**2+ (16.5/1e4)*model.T[i] + 0.793)
# 				#Vstar2 = model.d + model.e*X_mea
#
# 				#V_sln = X_mea*V_mea + X_H2O*V_H2O + X_CO2*model.V_CO2 + X_mea*X_H2O*model.Vstar # + X_mea*X_CO2*Vstar2
# 				V_sln = X_mea*V_mea + X_H2O*V_H2O + X_CO2*V_CO2 + X_mea*X_H2O*model.Vstar # + X_mea*X_CO2*Vstar2
#
# 				expr+= (model.rho_sln[i] - MW_sln/V_sln)**2
# 			return expr;
#
# 		model.OBJ = Objective(rule=obj_expression)
#
# 	def alamo_datasets(self,alamo_model):
# 		# Measured value of interest
# 		zdata = np.asarray(self.data.get_column('Density Data').values);
# 		self.alamo_model.zdata = zdata.reshape((len(zdata),1))
#
# 		# Molecular Weight
# 		MW_mea = 61.08
# 		MW_H2O = 18.02
# 		MW_CO2 = 44.01
#
# 		xdata = []
# 		xdata.append(self.data.get_column('T').values)
#
# 		alpha = self.data.get_column('CO2 Loading').values;
# 		rstar = self.data.get_column('r').values;
#
# 		X_mea = [1/(1+ a + (MW_mea/MW_H2O)*((1-r)/r)) for a,r in zip(alpha, rstar)];
# 		X_CO2 = [X_m* a for X_m,a in zip(X_mea,alpha)]
# 		X_H2O =  [1 - X_m - X_C for X_m, X_C in zip(X_mea, X_CO2)];
#
# 		xdata.append(X_mea);
# 		xdata.append(X_CO2);
# 		xdata.append(X_H2O);
#
# 		self.alamo_model.xdata = np.transpose(xdata)
# 		# self.alamo_model.enumerate_datasets(4, ['T','y[\'MEA\']','y[\'CO2\']', 'y[\'H2O\']'], self.alamo_model.xdata)
# 		self.alamo_model.enumerate_datasets(4, ['T','yMEA','yCO2', 'yH2O'], self.alamo_model.xdata)
# 		# self.alamo_model.enumerate_datasets(4,'T yMEA yCO2 yH2O', self.alamo_model.xdata)
#
# 	def alamo_output(self, alamo_model,textfile):
# 		textfile.write("def _V_liq(self): \n");
# 		textfile.write("\t# ALAMO Model\n")
# 		textfile.write("\t(T, P, y) = self.deref_tpy()\n");
# 		textfile.write("\tVst = -1.8218e-6\n")
# 		textfile.write("\tmw = PhysPropParams.mw\n")
# 		textfile.write("\tr = self.wt_frac[\"MEA\"]\n\n");
# 		expr = "sum(y[i]*mw[i]/(a[i]*T**2 + b[i]*T + c[i]) for i in y.keys()) + y['MEA']*y['H2O']*Vst)"
# 		textfile.write("\tself.var_or_expr(name=\"visc_liq\", doc=\"Liquid molar volume\", expr=%s)\n\n" %alamo_model['model'])
#
# class SurfaceTensionModel(PropertyModel):
#
# 	# class SurfaceTensionParams(object):
# 		#     ### Surface tension parameters
# 		#     ###
# 		#     # Pure component surface tension for MEA and H2O
# 		#     # (Asprion 2005)
# 		#     C = {'MEA':(0.09945, 1.067,  0,      0    ),
# 		#          'H2O':(0.18548, 2.717, -3.554, 2.047)}
# 		#     # Parameters for CO2 surface tension
# 		#     S = {'CO2':(-5.987, 3.7699, -0.43164, 0.018155, -0.01207, 0.002119)}
# 		#     F = {
# 		#         'a':  2.4558,
# 		#         'b': -1.5311,
# 		#         'c':  3.4994,
# 		#         'd': -5.6398,
# 		#         'e': 10.2109,
# 		#         'f':  2.3122,
# 		#         'g':  4.5608,
# 		#         'h': -2.3924,
# 		#         'i':  5.3324,
# 		#         'j':-12.0494}
#
#
# 	    # def _surf_tens(self):
# 	     #    (T, P, y) = self.deref_tpy()
# 	     #    if self.external:
# 	     #        self.var_or_expr(name="surf_tens",
# 	     #            expr=self.extf_sig(T, y['CO2'], y['H2O'], y['MEA']))
# 	     #        return
# 	     #    F = SurfaceTensionParams.F
# 	     #    Tc = PhysPropParams.Tc
# 	     #    c = SurfaceTensionParams.C
# 	     #    S = SurfaceTensionParams.S
# 	     #    # Variables
# 	     #    r = self.wt_frac['CO2']
# 	     #    alpha = self.expr_loading = Expression(expr=y["CO2"]/y["MEA"])
#
# 	     #    # Constraints
# 	     #    def rule_sigma_i(blk, i):
# 	     #        if i in ['MEA', 'H2O']: #correlation type 1
# 	     #            return c[i][0]*(1-T/Tc[i])**\
# 	     #                (c[i][1] + c[i][2]*T/Tc[i] + c[i][3]*(T/Tc[i])**2)
# 	     #        elif i=='CO2': #correlation type 2
# 	     #            return S['CO2'][0]*r**2 + S['CO2'][1]*r + S['CO2'][2] +\
# 	     #                T*(S['CO2'][3]*r**2 + S['CO2'][4]*r+S['CO2'][5])
# 	     #    self.var_or_expr(["MEA", "CO2", "H2O"], name="sigma_i",
# 	     #        rule=rule_sigma_i)
# 	     #    sigma_i = self.sigma_i
# 	     #    self.var_or_expr(name="surf_tens", expr=sigma_i['H2O'] +
# 	     #        (sigma_i['H2O'] - sigma_i['CO2'])*y['CO2']*
# 	     #        (F['a']+F['b']*alpha+F['c']*alpha**2+F['d']*r+F['e']*r**2) +
# 	     #        (sigma_i['MEA'] - sigma_i['H2O'])*y['MEA']*
# 	     #        (F['f']+F['g']*alpha+F['h']*alpha**2+F['i']*r+F['j']*r**2))
#
#
# 	def __init__(self, data, metadata, outputFile, almtrace, description):
# 		PropertyModel.__init__(self, data, metadata, outputFile, almtrace, description);
#
#
# 	def results_str(self, model):
# 		# print self.results;
# 		print model.a.value, model.b.value, model.c.value, model.d.value, model.e.value, model.f.value, model.g.value, model.h.value, model.i.value, model.j.value
# 		print "Sum of squared error = %f" % model.OBJ()
#
#
# 	def model_output(self, model, textfile):
# 		textfile.write("def _surf_tens(self): \n");
# 		textfile.write("\t# Pure component surface tension for MEA and H2O\n")
# 		textfile.write("\t# (Asprion 2005)\n");
# 		textfile.write("\t(T, P, y) = self.deref_tpy()\n")
# 		textfile.write("\tC = {'MEA':(0.09945, 1.067,  0,      0    ),\n")
# 		textfile.write("\t'H2O':(0.18548, 2.717, -3.554, 2.047)}\n\n");
#
# 		textfile.write("\t# Parameters for CO2 surface tension\n")
# 		textfile.write("\tS = {'CO2':(-5.987, 3.7699, -0.43164, 0.018155, -0.01207, 0.002119)}\n");
# 		textfile.write("\tF = {")
# 		for v in model.component_objects(Var, active = True):
# 			if v.name== 'j':
# 				textfile.write("\n\t'%s' = %s}\n" %(v.name, v.value))
# 			else:
# 				textfile.write("\n\t'%s' = %s," %(v.name, v.value))
# 		textfile.write("\tTc = PhysPropParams.Tc\n")
#
# 		textfile.write("\t# Variables\n")
# 		textfile.write("\tr = self.wt_frac['CO2']\n");
# 		textfile.write("\talpha = self.expr_loading = Expression(expr=y[\"CO2\"]/y[\"MEA\"])\n");
#
#
# 		textfile.write("\t# Constraints\n");
# 		textfile.write("\tdef rule_sigma_i(blk, i):\n");
# 		textfile.write("\t\tif i in ['MEA', 'H2O']: #correlation type 1\n");
# 		textfile.write("\t\t\treturn c[i][0]*(1-T/Tc[i])** (c[i][1] + c[i][2]*T/Tc[i] + c[i][3]*(T/Tc[i])**2)\n");
# 		textfile.write("\t\telif i=='CO2': #correlation type 2\n");
# 		textfile.write("\t\t\treturn S['CO2'][0]*r**2 + S['CO2'][1]*r + S['CO2'][2] + T*(S['CO2'][3]*r**2 + S['CO2'][4]*r+S['CO2'][5])\n");
# 		textfile.write("\tself.var_or_expr([\"MEA\", \"CO2\", \"H2O\"], name=\"sigma_i\", rule=rule_sigma_i)\n");
# 		textfile.write("\tsigma_i = self.sigma_i\n");
# 		textfile.write("\tself.var_or_expr(name=\"surf_tens\", expr=sigma_i['H2O'] +(sigma_i['H2O'] - sigma_i['CO2'])*y['CO2']*(F['a']+F['b']*alpha+F['c']*alpha**2+F['d']*r+F['e']*r**2) +(sigma_i['MEA'] - sigma_i['H2O'])*y['MEA']*(F['f']+F['g']*alpha+F['h']*alpha**2+F['i']*r+F['j']*r**2))\n\n");
#
# 	def data_sets(self):
# 		self.I = np.asarray(range(self.data.num_rows));
#
# 		self.Temp = dict(zip(self.I,self.data.get_column('T').values));
# 		self.alpha = dict(zip(self.I,self.data.get_column('CO2 Loading').values));
# 		self.rstar = dict(zip(self.I,self.data.get_column('r').values));
#
# 		# Measured value of interest
# 		self.yVal = dict(zip(self.I,self.data.get_column('Surface Tension').values));
#
# 	def sets(self, model):
# 		model.I = Set(initialize=self.I)
#
# 	def parameters(self, model):
# 		# Given parameters of the system
# 		model.alpha = Param(model.I, initialize=self.alpha)
# 		model.rstar = Param(model.I, initialize=self.rstar)
# 		model.T = Param(model.I, initialize=self.Temp)
#
# 		# Measured value of interest
# 		model.y = Param(model.I, initialize=self.yVal)
#
# 	def variables(self, model):
# 		model.a = Var(domain=Reals, initialize = 0)
# 		model.b = Var(domain= Reals, initialize = 0)
# 		model.c = Var(domain=Reals, initialize = 0)
# 		model.d = Var(domain=Reals, initialize = 0)
# 		model.e = Var(domain=Reals, initialize = 0)
# 		model.f = Var(domain=Reals, initialize = 0)
# 		model.g = Var(domain=Reals, initialize = 0)
# 		model.h = Var(domain=Reals, initialize = 0)
# 		model.i = Var(domain=Reals, initialize = 0)
# 		model.j = Var(domain=Reals, initialize = 0)
#
# 	def objective(self, model):
#
# 		def obj_expression(model):
# 			expr = 0
# 			MW_mea = 61.08
# 			MW_h2o = 18.02
# 			MW_co2 = 44.01
#
# 			Tc_h2o = 647.13
# 			Tc_mea = 614.45
#
# 			c1_h2o = 0.18548
# 			c2_h2o = 2.717
# 			c3_h2o = -3.554
# 			c4_h2o = 2.047
# 			c1_mea = 0.09945
# 			c2_mea = 1.067
# 			c3_mea = 0
# 			c4_mea = 0
#
# 			S1 = -5.987
# 			S2 = 3.7699
# 			S3 = -0.43164
# 			S4 = 0.018155
# 			S5 = -0.01207
# 			S6 = 0.002119
#
# 			for i in model.I:
# 				X_mea = 1/(1+model.alpha[i]+(MW_mea/MW_h2o)*(1-model.rstar[i])/model.rstar[i])
# 				X_co2 = model.alpha[i]*X_mea
#
# 				sigma_h2o = c1_h2o*(1-model.T[i]/Tc_h2o)**(c2_h2o+c3_h2o*(model.T[i]/Tc_h2o)+c4_h2o*(model.T[i]/Tc_h2o)**2)
# 				sigma_mea = c1_mea*(1-model.T[i]/Tc_mea)**(c2_mea+c3_mea*(model.T[i]/Tc_mea)+c4_mea*(model.T[i]/Tc_mea)**2)
# 				sigma_co2 = S1*model.rstar[i]**2+S2*model.rstar[i]+S3+model.T[i]*(S4*model.rstar[i]**2+S5*model.rstar[i]+S6)
#
# 				f_func = model.a+model.b*model.alpha[i]+model.c*model.alpha[i]**2+model.d*model.rstar[i]+model.e*model.rstar[i]**2
# 				g_func = model.f+model.g*model.alpha[i]+model.h*model.alpha[i]**2+model.i*model.rstar[i]+model.j*model.rstar[i]**2
#
# 				sigma_calc=sigma_h2o+(sigma_co2-sigma_h2o)*f_func*X_co2+(sigma_mea-sigma_h2o)*g_func*X_mea
#
# 				expr+= (model.y[i] - sigma_calc)**2
#
# 			return expr
#
# 		model.OBJ = Objective(rule=obj_expression)
#
# 	def alamo_datasets(self,alamo_model):
# 		# Measured value of interest
# 		zdata = np.asarray(self.data.get_column('Surface Tension').values);
# 		#print zdata
# 		self.alamo_model.zdata = zdata.reshape((len(zdata),1))
#
# 		MW_mea = 61.08
# 		MW_h2o = 18.02
# 		MW_co2 = 44.01
#
# 		Tc_h2o = 647.13
# 		Tc_mea = 614.45
#
# 		c1_h2o = 0.18548
# 		c2_h2o = 2.717
# 		c3_h2o = -3.554
# 		c4_h2o = 2.047
# 		c1_mea = 0.09945
# 		c2_mea = 1.067
# 		c3_mea = 0
# 		c4_mea = 0
#
# 		S1 = -5.987
# 		S2 = 3.7699
# 		S3 = -0.43164
# 		S4 = 0.018155
# 		S5 = -0.01207
# 		S6 = 0.002119
#
# 		xdata = []
# 		T = self.data.get_column('T').values;
# 		alpha = self.data.get_column('CO2 Loading').values;
# 		rstar = self.data.get_column('r').values;
#
# 		X_mea = [1/(1+a+(MW_mea/MW_h2o)*(1-r)/r) for a,r in zip(alpha, rstar)]
# 		X_co2 = [a*x_m for a, x_m in zip(alpha, X_mea)]
#
# 		xdata.append(self.data.get_column('T').values)
# 		xdata.append(X_mea)
# 		xdata.append(X_co2)
#
# 		sigma_h2o = [c1_h2o*(1-t/Tc_h2o)**(c2_h2o+c3_h2o*(t/Tc_h2o)+c4_h2o*(t/Tc_h2o)**2) for t in T]
# 		sigma_mea = [c1_mea*(1-t/Tc_mea)**(c2_mea+c3_mea*(t/Tc_mea)+c4_mea*(t/Tc_mea)**2) for t in T]
# 		sigma_co2 = [S1*r**2+S2*r+S3+t*(S4*r**2+S5*r+S6) for r,t in zip(rstar, T)]
#
# 		xdata.append(sigma_h2o)
# 		xdata.append(sigma_mea)
# 		xdata.append(sigma_co2)
#
# 		xdata.append(rstar)
#
# 		self.alamo_model.xdata = np.transpose(xdata)
# 		#self.alamo_model.enumerate_datasets(4, ['T','y[\'MEA\']','y[\'CO2\']', 'sigma_i[\'H2O\']', 'sigma_i[\'MEA\']', 'sigma_i[\'CO2\']', 'r'], self.alamo_model.xdata)
# 		self.alamo_model.enumerate_datasets(4, ['T','yM','yC', 'sigma_iH2O', 'sigma_iMEA', 'sigma_iCO2', 'r'], self.alamo_model.xdata)
#
# 	def alamo_output(self, alamo_model, textfile):
# 		textfile.write("def _surf_tens(self): \n");
# 		textfile.write("\t# ALAMO Model\n")
# 		textfile.write("\t# Pure component surface tension for MEA and H2O\n")
# 		textfile.write("\t# (Asprion 2005)\n");
# 		textfile.write("\t(T, P, y) = self.deref_tpy()\n")
# 		textfile.write("\tC = {'MEA':(0.09945, 1.067,  0,      0    ),\n")
# 		textfile.write("\t'H2O':(0.18548, 2.717, -3.554, 2.047)}\n\n");
#
# 		textfile.write("\t# Parameters for CO2 surface tension\n")
# 		textfile.write("\tS = {'CO2':(-5.987, 3.7699, -0.43164, 0.018155, -0.01207, 0.002119)}\n");
# 		textfile.write("\tF = {")
# 		# for v in model.component_objects(['T','y[\'MEA\']','y[\'CO2\']', 'sigma_i[\'H2O\']', 'sigma_i[\'MEA\']', 'sigma_i[\'CO2\']'], active = True):
# 		# 	if v.name== 'j':
# 		# 		textfile.write("\n\t'%s' = %s}\n" %(v.name, v.value))
# 		# 	else:
# 		# 		textfile.write("\n\t'%s' = %s," %(v.name, v.value))
# 		textfile.write("}\n")
# 		textfile.write("\tTc = PhysPropParams.Tc\n")
#
# 		textfile.write("\t# Variables\n")
# 		textfile.write("\tr = self.wt_frac['CO2']\n");
# 		textfile.write("\talpha = self.expr_loading = Expression(expr=y[\"CO2\"]/y[\"MEA\"])\n");
#
#
# 		textfile.write("\t# Constraints\n");
# 		textfile.write("\tdef rule_sigma_i(blk, i):\n");
# 		textfile.write("\t\tif i in ['MEA', 'H2O']: #correlation type 1\n");
# 		textfile.write("\t\t\treturn c[i][0]*(1-T/Tc[i])** (c[i][1] + c[i][2]*T/Tc[i] + c[i][3]*(T/Tc[i])**2)\n");
# 		textfile.write("\t\telif i=='CO2': #correlation type 2\n");
# 		textfile.write("\t\t\treturn S['CO2'][0]*r**2 + S['CO2'][1]*r + S['CO2'][2] + T*(S['CO2'][3]*r**2 + S['CO2'][4]*r+S['CO2'][5])\n");
# 		textfile.write("\tself.var_or_expr([\"MEA\", \"CO2\", \"H2O\"], name=\"sigma_i\", rule=rule_sigma_i)\n");
# 		textfile.write("\tsigma_i = self.sigma_i\n");
# 		textfile.write("\tself.var_or_expr(name=\"surf_tens\", expr=%s)\n\n" %alamo_model['model'].replace("^","**"));
#
# class ViscosityModel(PropertyModel):
#
# 	# class ViscLiqParams(object):
# 		#     """
# 		#     Viscosity parameters for an MEA solvent
# 		#     """
# 		#     a = -0.0838
# 		#     b = 2.8817
# 		#     c = 33.651
# 		#     d = 1817.0
# 		#     e = 0.00847
# 		#     f = 0.0103
# 		#     g = -2.3890
#
#
# 	    # def _visc_liq(self):
# 	    #     (T, P, y) = self.deref_tpy()
# 	    #     if self.external:
# 	    #         self.var_or_expr(name="visc_liq",
# 	    #             expr=self.extf_mu_liq(T, y['CO2'], y['MEA']))
# 	    #         return
# 	    #     a = ViscLiqParams.a
# 	    #     b = ViscLiqParams.b
# 	    #     c = ViscLiqParams.c
# 	    #     d = ViscLiqParams.d
# 	    #     e = ViscLiqParams.e
# 	    #     f = ViscLiqParams.f
# 	    #     g = ViscLiqParams.g
# 	    #     r = self.wt_frac["MEA"]
#
# 	    #     self.var_or_expr(name="visc_liq",
# 	    #         expr=1.002e-3*10**((1.3272*(293.15 - T - 0.001053*(T-293.15)**2))/
# 	    #         (T-168.15))*exp(100*r*(T*(a*100*r + b) + c*100*r + d)*
# 	    #         (y["CO2"]/y["MEA"]*(e*100*r + f*T + g) + 1)/T**2))
#
# 	def __init__(self, data, metadata, outputFile, almtrace, description):
# 		PropertyModel.__init__(self, data, metadata, outputFile, almtrace, description);
#
#
# 	def results_str(self, model):
# 		print model.a.value, model.b.value, model.c.value, model.d.value, model.e.value, model.f.value, model.g.value;
# 		print "Sum of squared error = %f" % model.OBJ()
#
# 	def model_output(self, model, textfile):
# 		textfile.write("def _visc_liq(self): \n");
# 		textfile.write("\t(T, P, y) = self.deref_tpy()\n");
# 		for v in model.component_objects(Var, active = True):
# 			textfile.write("\t%s = %s\n" %(v.name, v.value))
# 		textfile.write("\tr = self.wt_frac[\"MEA\"]\n\n");
# 		expr = "1.002e-3*10**((1.3272*(293.15 - T - 0.001053*(T-293.15)**2))/(T-168.15))*exp(100*r*(T*(a*100*r + b) + c*100*r + d)*(y[\"CO2\"]/y[\"MEA\"]*(e*100*r + f*T + g) + 1)/T**2)"
# 		textfile.write("\tself.var_or_expr(name=\"visc_liq\", expr=%s)\n\n" %expr)
#
#
# 	def data_sets(self):
# 		self.I = np.asarray(range(self.data.num_rows));
# 		W_mea_vals = [v*100 for v in self.data.get_column('r').values]
# 		self.W_mea = dict(zip(self.I, W_mea_vals));
# 		self.Temp = dict(zip(self.I,self.data.get_column('T').values));
# 		self.alpha = dict(zip(self.I,self.data.get_column('CO2 Loading').values));
#
# 		# Measured value of interest
# 		self.u_sln = dict(zip(self.I,self.data.get_column('Viscosity Value').values));
#
# 	def sets(self, model):
# 		model.I = Set(initialize=self.I)
#
# 	def parameters(self, model):
# 		# Given parameters of the system
# 		model.alpha = Param(model.I, initialize=self.alpha)
# 		model.W_mea = Param(model.I, initialize=self.W_mea)
# 		model.T = Param(model.I, initialize=self.Temp)
#
# 		# Measured value of interest
# 		model.u_sln = Param(model.I, initialize=self.u_sln)
#
# 	def variables(self, model):
# 		model.a = Var(domain=Reals, initialize = 0)
# 		model.b = Var(domain= Reals, initialize = 0)
# 		model.c = Var(domain=Reals, initialize = 0)
# 		model.d = Var(domain=Reals, initialize = 0)
# 		model.e = Var(domain=Reals, initialize = 0)
# 		model.f = Var(domain=Reals, initialize = 0)
# 		model.g = Var(domain=Reals, initialize = 0)
#
# 	def objective(self, model):
#
# 		def obj_expression(model):
# 				expr = 0;
# 				for i in model.I:
# 					u_H2O = 1.002*10**((1.3272*(293.15 - model.T[i] - 0.001053*(model.T[i]-293.15)**2))/(model.T[i]-168.15))
# 					expr+= (model.u_sln[i] - u_H2O*exp((((model.a*model.W_mea[i]+model.b)*model.T[i] + model.c*model.W_mea[i] + model.d)*(model.alpha[i]*(model.e*model.W_mea[i] + model.f*model.T[i] + model.g) + 1)* model.W_mea[i])/(model.T[i]**2)))**2
# 				return expr
#
# 		model.OBJ = Objective(rule=obj_expression)
#
# 	def alamo_datasets(self,alamo_model):
# 		# Measured value of interest
# 		zdata = np.asarray(self.data.get_column('Viscosity Value').values);
# 		zdata = np.log(zdata)
# 		#print zdata
# 		self.alamo_model.zdata = zdata.reshape((len(zdata),1))
#
# 		xdata = []
# 		xdata.append(self.data.get_column('T').values)
# 		xdata.append(self.data.get_column('CO2 Loading').values)
# 		W_mea_vals = [v*100 for v in self.data.get_column('r').values]
# 		xdata.append(W_mea_vals)
#
# 		u_H2O = [1.002*10**((1.3272*(293.15 - t - 0.001053*(t-293.15)**2))/(t-168.15)) for t in self.data.get_column('T').values]
# 		# xdata.append(u_H2O)
#
# 		self.alamo_model.xdata = np.transpose(xdata)
# 		self.alamo_model.enumerate_datasets(3, ['T','a','W'], self.alamo_model.xdata)
# 		# self.alamo_model.enumerate_datasets(3, ['T','y[\"CO2\"]/y[\"MEA\"]','r'], self.alamo_model.xdata)
#
# 	def alamo_output(self, alamo_model, textfile):
# 		textfile.write("def _visc_liq(self): \n");
# 		textfile.write("\t# ALAMO Model\n")
# 		textfile.write("\t(T, P, y) = self.deref_tpy()\n");
# 		# for v in model.component_objects(Var, active = True):
# 		# 	textfile.write("\t%s = %s\n" %(v.name, v.value))
# 		textfile.write("\tr = self.wt_frac[\"MEA\"]\n\n");
# 		expr = "1.002e-3*10**((1.3272*(293.15 - T - 0.001053*(T-293.15)**2))/(T-168.15))*exp(100*r*(T*(a*100*r + b) + c*100*r + d)*(y[\"CO2\"]/y[\"MEA\"]*(e*100*r + f*T + g) + 1)/T**2)"
# 		textfile.write("\tself.var_or_expr(name=\"visc_liq\", expr=%s)\n\n" %alamo_model['model'])
#
# # for method_name in dir(EnthalpyModel):
# # 	print method_name
# # help(EnthalpyModel.alamo_datasets)
#
# # SteamPropertyPackage()
#
# MEAPropertyPackage()
#
#
#
#
# ########### SNIPPETS
#
# 	# def variables(self, model):
# 	# 	# Regressed parameter values
# 	# 	model.a_mea = Var(domain=Reals, initialize = 1)
# 	# 	model.b_mea = Var(domain= Reals, initialize = 1)
# 	# 	model.c_mea = Var(domain=Reals, initialize = 1)
#
# 	# 	#model.a_CO2 = Var(domain=Reals, initialize = 0)
# 	# 	#model.b_CO2 = Var(domain= Reals, initialize = 0)
# 	# 	model.c_CO2 = Var(domain=Reals, initialize = 1)
#
# 	# 	#model.d = Var(domain=Reals, initialize = 1)
# 	# 	#model.e = Var(domain=Reals, initialize = 1)
# 	# 	#model.V_CO2 = Var(domain=NonNegativeReals, initialize = 1)
# 	# 	model.Vstar = Var(domain=Reals, initialize = -1)
#
# 	# def objective(self, model):
#
# 	# 	def obj_expression(model):
# 	# 		expr = 0
#
# 	# 		# Molecular Weight
# 	# 		MW_mea = 61.08
# 	# 		MW_H2O = 18.02
# 	# 		MW_CO2 = 44.01
#
# 	# 		for i in model.I:
#
# 	# 			# Mol Fractions
# 	# 			X_mea = 1/(1+ model.alpha[i] + (MW_mea/MW_H2O)*((1-model.r[i])/model.r[i]))
# 	# 			X_CO2 = X_mea * model.alpha[i]
# 	# 			X_H2O =  1 - X_mea - X_CO2
# 	# 			MW_sln = X_mea*MW_mea + X_CO2*MW_CO2 + X_H2O*MW_H2O
#
# 	# 			# Volume
# 	# 			V_mea = MW_mea/((model.a_mea/1e7)*model.T[i]**2 + (model.b_mea/1e4)*model.T[i] + model.c_mea)
# 	# 			V_CO2 = MW_CO2/model.c_CO2;
# 	# 			#V_CO2 = MW_CO2/((model.a_CO2/1e7)*model.T[i]**2 + (model.b_CO2/1e4)*model.T[i] + model.c_CO2)
# 	# 			V_H2O = MW_H2O/(-(32.484/1e7)*model.T[i]**2+ (16.5/1e4)*model.T[i] + 0.793)
# 	# 			#Vstar2 = model.d + model.e*X_mea
#
# 	# 			#V_sln = X_mea*V_mea + X_H2O*V_H2O + X_CO2*model.V_CO2 + X_mea*X_H2O*model.Vstar # + X_mea*X_CO2*Vstar2
# 	# 			V_sln = X_mea*V_mea + X_H2O*V_H2O + X_CO2*V_CO2 + X_mea*X_H2O*model.Vstar # + X_mea*X_CO2*Vstar2
#
# 	# 			expr+= (model.rho_sln[i] - MW_sln/V_sln)**2
# 	# 		return expr;
#
# 	# 	model.OBJ = Objective(rule=obj_expression)
#



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Modeling for saturated densities and vapor pressure
"""

import sys
from idaes.surrogate import alamopy
import importlib

from . import DataImport
from . import DataManipulation


molecule = ""
max_time = 1000



global DLfun, DVfun, PVfun
DLfun, DVfun, PVfun = None, None, None

DataToWrite = []


def DL():
    "ALAMO regression of Saturated Liquid Density"
    global DLfun
    DataImport.DL(molecule)
    Values = DataImport.DLValues

    xval, zval = [], []
    for x in Values:
        manVal = DataManipulation.DL(x)
        xval.append(manVal[0])
        zval.append(manVal[1])

    res = alamopy.alamo(
        xval,
        zval,
        zlabels=["%sDL" % molecule],
        almname="%sDL" % molecule,
        monomialpower=(-1, 1, 2, 3),
        expfcns=1,
        savetrace=True,
        savescratch=True,
    )

    if res is None:
        raise Exception("Model does not exist for Saturated Liquid Density")

    DLfun = importlib.import_module("%sDL" % molecule, "f")


def getDL():
    """ 
    Imports the regressed saturated liquid density function
    """
    global DLfun
    if DLfun is None:
        DLfun = importlib.import_module("%sDL" % molecule, "f")
    return DLfun


def DV():
    """
    ALAMO regression of saturated vapor density
    """
    global DVfun
    DataImport.DV(molecule)
    Values = DataImport.DVValues

    xval, zval = [], []
    for x in Values:
        manVal = DataManipulation.DV(x)
        xval.append(manVal[0])
        zval.append(manVal[1])

    res = alamopy.alamo(
        xval,
        zval,
        zlabels=["%sDV" % molecule],
        almname="%sDV" % molecule,
        monomialpower=(1, 2, 3, 4, 5, 6),
        savetrace=True,
    )
    # print(res['model'], res['ssr'], res['R2'])

    if res is None:
        raise Exception("Model does not exist for Saturated Vapor Density")

    DVfun = importlib.import_module("%sDV" % molecule, "f")


def getDV():
    """ 
    Imports the regressed saturated vapor density function
    """
    global DVfun
    if DVfun is None:
        DVfun = importlib.import_module("%sDV" % molecule, "f")
    return DVfun


def PV():
    "ALAMO regression of vapor pressure"
    global PVfun
    DataImport.PV(molecule)
    Values = DataImport.Values

    xval, zval = [], []
    for x in Values:
        manVal = DataManipulation.PV(x)
        xval.append([manVal[0], manVal[1]])
        zval.append(manVal[2])

    res = alamopy.alamo(
        xval,
        zval,
        zlabels=["%sPV" % molecule],
        almname="%sPV" % molecule,
        monomialpower=(1, 2, 3, 4, 5, 6),
    )
    # print(res['model'], res['ssr'], res['R2'])

    if res is None:
        raise Exception("Model does not exist for Vapor Pressure")

    PVfun = importlib.import_module("%sPV" % molecule, "f")


def getPV():
    """ 
    Imports the regressed vapor pressure function
    """
    global PVfun
    if PVfun is None:
        PVfun = importlib.import_module("%sPV" % molecule, "f")
    return PVfun



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
This provides valve models for steam and liquid water.  These are for
steam cycle control valves and the turbine throttle valves.
"""

__Author__ = "John Eslick"

from enum import Enum

from pyomo.common.config import ConfigBlock, ConfigValue, In, ConfigList
from pyomo.environ import (Var, Expression, SolverFactory, value,
                           Constraint, sqrt, Param)
from pyomo.opt import TerminationCondition

from idaes.core import declare_process_block_class
from idaes.unit_models.pressure_changer import (
    PressureChangerData,
    ThermodynamicAssumption,
    MaterialBalanceType,
)
from idaes.core.util import from_json, to_json, StoreSpec
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.logger import getIdaesLogger, getInitLogger, init_tee, condition

_log = getIdaesLogger(__name__)


class ValveFunctionType(Enum):
    linear = 1
    quick_opening = 2
    equal_percentage = 3
    custom = 4


def _define_config(config):
    config.compressor = False
    config.get('compressor')._default = False
    config.get('compressor')._domain = In([False])
    config.material_balance_type = MaterialBalanceType.componentTotal
    config.get('material_balance_type')._default = \
        MaterialBalanceType.componentTotal
    config.thermodynamic_assumption = ThermodynamicAssumption.adiabatic
    config.get('thermodynamic_assumption')._default = \
        ThermodynamicAssumption.adiabatic
    config.get('thermodynamic_assumption')._domain = \
        In([ThermodynamicAssumption.adiabatic])
    config.declare("valve_function", ConfigValue(
        default=ValveFunctionType.linear,
        domain=In(ValveFunctionType),
        description="Valve function type, if custom provide an expression rule",
        doc="""The type of valve function, if custom provide an expression rule
with the valve_function_rule argument.
**default** - ValveFunctionType.linear
**Valid values** - {
ValveFunctionType.linear,
ValveFunctionType.quick_opening,
ValveFunctionType.equal_percentage,
ValveFunctionType.custom}"""))
    config.declare("valve_function_rule", ConfigValue(
        default=None,
        description="This is a rule that returns a time indexed valve function expression.",
        doc="""This is a rule that returns a time indexed valve function expression.
This is required only if valve_function==ValveFunctionType.custom"""))
    config.declare("phase", ConfigValue(
        default="Vap",
        domain=In(("Vap", "Liq")),
        description='Expected phase of fluid in valve in {"Liq", "Vap"}'))


def _linear_rule(b, t):
    return b.valve_opening[t]

def _quick_open_rule(b, t):
    return sqrt(b.valve_opening[t])

def _equal_percentage_rule(b, t):
    return b.alpha**(b.valve_opening[t] - 1)

def _liquid_pressure_flow_rule(b, t):
    """
    For liquid F = Cv*sqrt(Pi**2 - Po**2)*f(x)
    """
    Po = b.control_volume.properties_out[t].pressure
    Pi = b.control_volume.properties_in[t].pressure
    F = b.control_volume.properties_in[t].flow_mol
    Cv = b.Cv
    fun = b.valve_function[t]
    return (1/b.flow_scale**2)*F**2 == (1/b.flow_scale**2)*Cv**2*(Pi - Po)*fun**2

def _vapor_pressure_flow_rule(b, t):
    """
    For vapor F = Cv*sqrt(Pi**2 - Po**2)*f(x)
    """
    Po = b.control_volume.properties_out[t].pressure
    Pi = b.control_volume.properties_in[t].pressure
    F = b.control_volume.properties_in[t].flow_mol
    Cv = b.Cv
    fun = b.valve_function[t]
    return (1/b.flow_scale**2)*F**2 == \
        (1/b.flow_scale**2)*Cv**2*(Pi**2 - Po**2)*fun**2


@declare_process_block_class("SteamValve", doc="Basic steam valve models")
class SteamValveData(PressureChangerData):
    # Same settings as the default pressure changer, but force to expander with
    # isentropic efficiency
    CONFIG = PressureChangerData.CONFIG()
    _define_config(CONFIG)

    def build(self):
        super().build()

        self.valve_opening = Var(self.flowsheet().config.time, initialize=1,
            doc="Fraction open for valve from 0 to 1")
        self.Cv = Var(initialize=0.1, doc="Valve flow coefficent, for vapor "
            "[mol/s/Pa] for liquid [mol/s/Pa^0.5]")
        self.flow_scale = Param(mutable=True, default=1e3, doc=
            "Scaling factor for pressure flow relation should be approximatly"
            " the same order of magnitude as the expected flow.")
        self.Cv.fix()
        self.valve_opening.fix()

        # set up the valve function rule.  I'm not sure these matter too much
        # for us, but the options are easy enough to provide.
        if self.config.valve_function == ValveFunctionType.linear:
            rule = _linear_rule
        elif self.config.valve_function == ValveFunctionType.quick_opening:
            rule = _quick_open_rule
        elif self.config.valve_function == ValveFunctionType.equal_percentage:
            self.alpha = Var(initialize=1, doc="Valve function parameter")
            self.alpha.fix()
            rule = equal_percentage_rule
        else:
            rule = self.config.valve_function_rule

        self.valve_function = Expression(self.flowsheet().config.time,
                                         rule=rule,
                doc="Valve function expression")

        if self.config.phase == "Liq":
            rule = _liquid_pressure_flow_rule
        else:
            rule = _vapor_pressure_flow_rule

        self.pressure_flow_equation = Constraint(self.flowsheet().config.time,
                                                 rule=rule)

    def initialize(self, state_args={}, outlvl=6, solver='ipopt',
        optarg={'tol': 1e-6, 'max_iter':30}):
        """
        Initialize the turbine stage model.  This deactivates the
        specialized constraints, then does the isentropic turbine initialization,
        then reactivates the constraints and solves.

        Args:
            state_args (dict): Initial state for property initialization
            outlvl : sets output level of initialization routine
                 * 0 = Use default idaes.init logger setting
                 * 1 = Maximum output
                 * 2 = Include solver output
                 * 3 = Return solver state for each step in subroutines
                 * 4 = Return solver state for each step in routine
                 * 5 = Final initialization status and exceptions
                 * 6 = No output
            solver (str): Solver to use for initialization
            optarg (dict): Solver arguments dictionary
        """
        # sp is what to save to make sure state after init is same as the start
        #   saves value, fixed, and active state, doesn't load originally free
        #   values, this makes sure original problem spec is same but initializes
        #   the values of free vars
        sp = StoreSpec.value_isfixed_isactive(only_fixed=True)
        istate = to_json(self, return_dict=True, wts=sp)

        self.deltaP[:].unfix()
        self.ratioP[:].unfix()

        # fix inlet and free outlet
        for t in self.flowsheet().config.time:
            for k, v in self.inlet.vars.items():
                v[t].fix()
            for k, v in self.outlet.vars.items():
                v[t].unfix()
            # to calculate outlet pressure
            Pout = self.outlet.pressure[t]
            Pin = self.inlet.pressure[t]
            if self.deltaP[t].value is not None:
                prdp = value((self.deltaP[t] - Pin)/Pin)
            else:
                prdp = -100 # crazy number to say don't use deltaP as guess
            if value(Pout/Pin) > 1 or value(Pout/Pin) < 0.0:
                if value(self.ratioP[t]) <= 1 and value(self.ratioP[t]) >= 0:
                    Pout.value = value(Pin*self.ratioP[t])
                elif prdp <= 1 and prdp >= 0:
                    Pout.value = value(prdp*Pin)
                else:
                    Pout.value = value(Pin*0.95)
            self.deltaP[t] = value(Pout - Pin)
            self.ratioP[t] = value(Pout/Pin)

        # Make sure the initialization problem has no degrees of freedom
        # This shouldn't happen here unless there is a bug in this
        dof = degrees_of_freedom(self)
        try:
            assert(dof == 0)
        except:
            _log.exception("degrees_of_freedom = {}".format(dof))
            raise

        # one bad thing about reusing this is that the log messages aren't
        # really compatible with being nested inside another initialization
        super().initialize(state_args=state_args,
            outlvl=outlvl, solver=solver, optarg=optarg)

        # reload original spec
        from_json(self, sd=istate, wts=sp)

    def _get_performance_contents(self, time_point=0):
        pc = super()._get_performance_contents(time_point=time_point)

        pc["vars"]["Opening"] = self.valve_opening[time_point]
        pc["vars"]["Valve Coefficient"] = self.Cv
        if self.config.valve_function == ValveFunctionType.equal_percentage:
            pc["vars"]["alpha"] = self.alpha

        pc["params"] = {}
        pc["params"]["Flow Scaling"] = self.flow_scale

        return pc



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Test doc build.

This does *not* try to build the docs. Instead it looks
for the log of the build and looks for errors in it.
"""
# stdlib
import logging
import os
import re

# third-party
import pytest


_log = logging.getLogger(__name__)


@pytest.fixture
def docs_path():
    """Find the docs.

    We have to assume that the docs are somewhere
    relative to either this file or the user's current
    working directory.
    Start with the working directory.

    Returns:
        str: Doc path or None if not found.
    """
    cwd = os.path.realpath(os.getcwd())
    # look for "./docs"
    path = os.path.join(cwd, "docs")
    if os.path.exists(path):
        return path
    # look for (last) "docs/" in full current working dir
    cwd_comp = cwd.split(os.path.sep)
    n = len(cwd_comp)
    try:
        idx = list(reversed(cwd_comp)).index("docs")
        return os.path.sep.join(cwd_comp[: n - idx])
    except ValueError:
        pass
    # look for docs relative to this file
    try:
        path = os.path.join(os.path.dirname(__file__), "..", "..", "docs")
        # look for {this file}/../../docs
        if os.path.exists(path):
            return path
        # look for "docs/" in file's full path
        file_comp = path.split(os.path.sep)
        n = len(file_comp)
        try:
            idx = list(reversed(file_comp)).index("docs")
            return os.path.sep.join(file_comp[: n - idx])
        except ValueError:
            pass
    except NameError:
        pass  # __file__ not defined(?)


ERRLOG = "sphinx-errors.txt"


def test_sphinx_build_log(docs_path):
    _log.info('docs path = "{}"'.format(docs_path))
    if docs_path is None:
        _log.warning('Could not find "docs" directory')
        return
    log_path = os.path.join(docs_path, ERRLOG)
    if not os.path.exists(log_path):
        _log.warning(
            'Could not find "{}" in docs directory: {}'.format(ERRLOG, log_path)
        )
        return
    log = open(log_path)
    for line in log:
        if "WARNING: " in line:
            pass
            # assert re.search(r'duplicate label sub(module|package)s', line), \
            #    'Non-trivial warning: {}'.format(line.strip())
        elif "ERROR: " in line:
            assert False, "Error: {}".format(line.strip())



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Functions for saving and loading Pyomo objects to json
"""

from pyomo.environ import *
from pyomo.network import Port, Arc
from pyomo.dae import *
from pyomo.core.base.component import ComponentData
import json
import datetime
import time
import gzip

# Some more inforation about this module
__author__ = "John Eslick"
__format_version__ = 3


def _set_active(o, d):
    """
    Set if component is active, used for read active attribute callback.
    Args:
        o: object whoes attribute is to be set
        d: attribute value
    Returns:
        None
    """
    if d:
        o.activate()
    else:
        o.deactivate()

def _set_fixed(o, d):
    """
    Set if variable is fixed, used for read fixed attribute callback.
    Args:
        o: object whoes attribute is to be set
        d: attribute value
    Returns:
        None
    """
    if d:
        o.fix()
    else:
        o.unfix()

def _get_value(o):
    """
    Get object value attribute callback.
    Args:
        o: object whoes attribute is to be set
        d: attribute value
    Returns:
        None
    """
    return value(o, exception=False)

def _set_value(o, d):
    """
    Set object value attribute callback. This doen't allow the value of an
    immutable paramter to be set (which would raise an exeption in Pyomo)
    Args:
        o: object whoes attribute is to be set
        d: attribute value
    Returns:
        None
    """
    if isinstance(o, Param) and not o._mutable:
        return #ignore requests to set immutable params
    else:
        try:
            o.value = d
        except AttributeError:
            o = d # this would be an indexed parameter

def _set_lb(o, d):
    """
    Set variable lower bound, used for read lb attribute callback.
    Args:
        o: object whoes attribute is to be set
        d: attribute value
    Returns:
        None
    """
    o.setlb(d)

def _set_ub(o, d):
    """
    Set variable upper bound, use for read ub attribute callback.
    Args:
        o: object whoes attribute is to be set
        d: attribute value
    Returns:
        None
    """
    o.setub(d)

def _only_fixed(o, d):
    """
    Returns a list of attributes to read for a variable, only whether it is
    fixed for non-fixed variables and if it is fixed and the value for fixed
    variables.  The allows you to set up a serializer that only reads fixed
    variable values.

    Args:
        o: Pyomo component being loaded
        d: State dictionary for the component o.
    Returns:
        An attribute list to read. Loads fixed for either fixed or un-fixed
        variables, but only reads in values for unfixed variables.  This is
        useful for intialization functions.
    """
    if d["fixed"]:
        return ("value", "fixed")
    else:
        return ("fixed",)

class Counter(object):
    """
    This is a counter object, which is an easy way to pass an interger pointer
    around between methods.
    """
    def __init__(self):
        self.count = 0

class StoreSpec(object):
    """
    A StoreSpec object tells the serializer functions what to read or write.
    The default settings will produce a StoreSpec configured to load/save the
    typical attributes required to load/save a model state.

    Args:
        classes: A list of classes to save.  Each class is represented by a
            list (or tupple) containing the following elements: (1) class
            (compared using isinstance) (2) attribute list or None, an emptry
            list store the object, but none of its attributes, None will not
            store objects of this class type (3) optional load filter function.
            The load filter function returns a list of attributes to read based
            on the state of an object and its saved state. The allows, for
            example, loading values for unfixed variables, or only loading
            values whoes current value is less than one. The filter function
            only applies to load not save. Filter functions take two arguments
            (a) the object (current state) and (b) the dictionary containing the
            saved state of an object.  More specific classes should come before
            more general classes.  For example if an obejct is a HeatExchanger
            and a UnitModel, and HeatExchanger is listed first, it will follow
            the HeatExchanger settings.  If UnitModel is listed first in the
            classes list, it will follow the UnitModel settings.
        data_classes: This takes the same form as the classes argument.
            This is for component data classes.
        skip_classes: This is a list of classes to skip.  If a class appears
            in the skip list, but also appears in the classes argument, the
            classes argument will override skip_classes. The use for this is to
            specifically exclude certain classes that would get caught by more
            general classes (e.g. UnitModel is in the class list, but you want
            to exclude HeatExchanger which is derived from UnitModel).
        ignore_missing: If True will ignore a component or attribute that exists
            in the model, but not in the stored state. If false an excpetion
            will be raised for things in the model that should be loaded but
            aren't in the stored state. Extra items in the stored state will not
            raise an exception regaurdless of this argument.
        suffix: If True store suffixes and component ids.  If false, don't store
            suffixes.
        suffix_filter: None to store all siffixes if suffix=True, or a list of
            suffixes to store if suffix=True
    """
    def __init__(
        self,
        classes=(
            (Param, ("_mutable",)),
            (Var, ()),
            (Component, ("active",))),
        data_classes=(
            (pyomo.core.base.var._VarData,
                ("fixed", "stale", "value", "lb", "ub")),
            (pyomo.core.base.param._ParamData, ("value",)),
            (int, ("value",)),
            (float, ("value",)),
            (pyomo.core.base.component.ComponentData, ("active",))),
        skip_classes=(ExternalFunction, Set, Port, Expression, RangeSet),
        ignore_missing=True,
        suffix=True,
        suffix_filter=None):
        """
        (see above)
        """
        # Callbacks are used for attributes that cannont be directly get or set
        self.write_cbs={ # Write callbacks (writing state so get attr)
            "value":_get_value}
        self.read_cbs={ # Reads callbacks (reading in state so set attr)
            "_mutable": None,
            "active":_set_active,
            "fixed":_set_fixed,
            "lb":_set_lb,
            "ub":_set_ub,
            "value":_set_value}
        # Add skip classes to classes list, with None as attr list to skip
        skip_classes2 = [] #need to put skips at front of list
        self.classes = [i[0] for i in classes]
        for i in skip_classes:
            if i not in self.classes:
                skip_classes2.append((i, None))
        classes = skip_classes2 + list(classes) # comined classes with skips
        # Create lists of classes, attribute lists, and filter functions
        # Can get class index from class list the use it to get associated items
        self.classes = [i[0] for i in classes]
        self.data_classes = [i[0] for i in data_classes]
        self.class_attrs = [i[1] for i in classes]
        self.data_class_attrs = [i[1] for i in data_classes]
        # Create filter function lists, use None if not supplied
        self.class_filter = []
        for i in classes:
            if len(i) < 3:
                self.class_filter.append(None)
            else:
                self.class_filter.append(i[2])
        self.data_class_filter = []
        for i in data_classes:
            if len(i) < 3:
                self.data_class_filter.append(None)
            else:
                self.data_class_filter.append(i[2])
        self.ignore_missing = ignore_missing
        self.include_suffix = suffix
        self.suffix_filter = suffix_filter

    def set_read_callback(self, attr, cb=None):
        """
        Set a callback to set an attribute, when reading from json or dict.
        """
        self.read_cbs[attr] = cb

    def set_write_callback(self, attr, cb=None):
        """
        Set a callback to get an attribute, when writing to json or dict.
        """
        self.write_cbs[attr] = cb

    def get_class_attr_list(self, o):
        """
        Look up what attributes to save/load for an Component object.
        Args:
            o: Object to look up attribute list for.
        Return:
            A list of attributes and a filter function for object type
        """
        alist = []  # Attributes to store
        ff = None   # Load filter function
        for i, cl in enumerate(self.classes):
            if isinstance(o, cl):
                alist = self.class_attrs[i]
                ff = self.class_filter[i]
                break
        return (alist, ff)

    def get_data_class_attr_list(self, o):
        """
        Look up what attributes to save/load for an ComponentData object.
        Args:
            o: Object to look up attribute list for.
        Return:
            A list of attributes and a filter function for object type
        """
        alist = []  # Attributes to store
        ff = None   # Load filter function
        for i, cl in enumerate(self.data_classes):
            if isinstance(o, cl):
                alist = self.data_class_attrs[i]
                ff = self.data_class_filter[i]
                break
        return (alist, ff)

    @classmethod
    def bound(cls):
        """Returns a StoreSpec object to store variable bounds only."""
        return cls(classes=((Var, ()),),
            data_classes=((pyomo.core.base.var._VarData, ("lb", "ub")),),
            suffix=False)

    @classmethod
    def value(cls):
        """Returns a StoreSpec object to store variable values only."""
        return cls(
            classes=((Var, ()),),
            data_classes=((pyomo.core.base.var._VarData, ("value",)),),
            suffix=False)

    @classmethod
    def isfixed(cls):
        """Returns a StoreSpec object to store if variables are fixed."""
        return cls(
            classes=((Var, ()),),
            data_classes=((pyomo.core.base.var._VarData, ("fixed",)),),
            suffix=False)

    @classmethod
    def suffix(cls, suffix_filter=None):
        return cls(
            classes=((Suffix, ()),),
            data_classes=(),
            suffix=True,
            suffix_filter=suffix_filter)

    @classmethod
    def value_isfixed(cls, only_fixed):
        """
        Return a StoreSpec object to store variable values and if fixed.

        Args:
            only_fixed: Only load fixed variable values
        """
        if only_fixed:
            return cls(
                classes=((Var, ()),),
                data_classes=(
                    (pyomo.core.base.var._VarData,
                        ("value", "fixed"), _only_fixed),),
                suffix=False)
        else:
            return cls(
                classes=((Var, ()),),
                data_classes=((pyomo.core.base.var._VarData,
                    ("value", "fixed")),),
                suffix=False)

    @classmethod
    def value_isfixed_isactive(cls, only_fixed):
        """
        Retur a StoreSpec object to store variable values, if variables are
        fixed and if components are active.

        Args:
            only_fixed: Only load fixed variable values
        """
        if only_fixed:
            return cls(
                classes=((Var, ()), (Param, ()), (Component, ("active",))),
                data_classes=(
                    (pyomo.core.base.var._VarData, ("value", "fixed"), _only_fixed),
                    (pyomo.core.base.param._ParamData, ("value",)),
                    (pyomo.core.base.component.ComponentData, ("active",))))
        else:
            return cls(
                classes=(
                    (Var, ()),
                    (Param, ()),
                    (Component, ("active",))),
                data_classes=(
                    (pyomo.core.base.var._VarData, ("value", "fixed")),
                    (pyomo.core.base.param._ParamData, ("value",)),
                    (pyomo.core.base.component.ComponentData, ("active",))))

def _may_have_subcomponents(o):
    """
    Args:
        o: an object.
    Returns:
        True if the object has a callable component_objects method, otherwise
        False.
    """
    if hasattr(o, "component_objects"):
        if hasattr(o.component_objects, "__call__"):
            return True

def _write_component(sd, o, wts, count=None, lookup={}, suffixes=[]):
    """
    Writes a component state to the save dictionary under a key given by the
    components name.

    Args:
        sd: dictionary to to save the object into, will create a key that is the
            object name (not fully qualified)
        o: object to save
        wts: a StoreSpec object indicating what object attributes to write
        count: count the number of Pyomo componets written also used for ids
        lookup: is a lookup table for compoent ids from components
        suffixes: is a list of suffixes, that we are delaying writing
    Returns:
        None
    """
    # Get list of attributes to save, also returns ff, which is a filter
    # function and only used in reading stuff back in.
    alist, ff = wts.get_class_attr_list(o)
    if alist is None: return #alist is none means skip this component type
    # Get the componet name, doesn't need to be fully quified or unique because
    # we are storing the state in a hierarchy structure
    oname = o.getname(fully_qualified=False)
    # Create a dictionary for this component, if storing suffixes assign it
    # a sequential id number and create a lookup table that takes the component
    # and returns its id for use later in writing suffix data
    sd[oname] = {"__type__":str(type(o))}
    if wts.include_suffix:
        sd[oname]["__id__"] = count.count
        lookup[id(o)] = count.count #used python id() here for efficency
    if count is not None: count.count += 1 # incriment the componet counter
    for a in alist: # store the desired attributes
        if a in wts.write_cbs:
            if wts.write_cbs[a] is None:
                sd[oname][a] = getattr(o, a, None)
            else:
                sd[oname][a] = wts.write_cbs[a](o)
        else:
            sd[oname][a] = getattr(o, a, None)
    sd[oname]["data"] = {} # create a dict for compoent data and subcomponents
    if isinstance(o, Suffix): # if is a suffix, make a list and delay writing
        if wts.include_suffix:        # data until all compoents have an assigned id
            if wts.suffix_filter is None or oname in wts.suffix_filter:
                suffixes.append(
                    {'sd':sd[oname]["data"], 'o':o, 'wts':wts, 'lookup':lookup})
    else: # if not suffix go ahead and write the data
        _write_component_data(sd=sd[oname]["data"], o=o, wts=wts, lookup=lookup,
                              count=count, suffixes=suffixes)

def _write_component_data(sd, o, wts, count=None, lookup={}, suffixes=[]):
    """
    Iterate through the component data and write to the sd dictionary. The keys
    for the data items are added to the dictionary. If the component has
    subcomponents they are written by a recursive call to _write_component under
    the __pyomo_components__ key.

    Args:
        sd: dictionary to to save the object into, will create keys that are the
            data object indexes repn.
        o: object to save
        wts: a StoreSpec object indicating what object attributes to write
        count: count the number of Pyomo componets written also used for ids
        lookup: is a lookup table for compoent ids from components
        suffixes: is a list of suffixes, that we are delaying writing
    Returns:
        None
    """
    if wts.include_suffix and isinstance(o, Suffix):
        # make special provision for writing suffixes.
        for key in o:
            el = o[key]
            sd[lookup[id(key)]] = el # Asssume keys are Pyomo model components
    else: # rest of compoents with normal componet data structure
        frst = True # on first item when true
        try:
            item_keys = o.keys()
        except AttributeError:
            item_keys = [None]
        for key in item_keys:
            if key is None and isinstance(o, ComponentData) \
                and not isinstance(o, Component):
                el = o
            else:
                el = o[key]
            if frst: # assume all item are same type, use first to get alist
                alist, ff = wts.get_data_class_attr_list(el) # get attributes
                if alist is None: return # if None then skip writing
            frst = False # done with first only stuff
            edict = {"__type__":str(type(el))}
            if wts.include_suffix: # if writing suffixes give data compoents an id
                edict["__id__"] = count.count
                lookup[id(el)] = count.count # and add to lookup table
            if count is not None: count.count += 1 # inciment component counter
            sd[repr(key)] = edict # stick item dict into component data dict
            for a in alist: # store desired attributes
                if a in wts.write_cbs:
                    if wts.write_cbs[a] is None:
                        edict[a] = getattr(el, a)
                    else:
                        edict[a] = wts.write_cbs[a](el)
                else:
                    edict[a] = getattr(el, a)
            hascomps = False # Has sub-components (like a Block would have)
            if _may_have_subcomponents(el): # block or block like component
                for o2 in el.component_objects(descend_into=False):
                    # loop through sub-components
                    if not hascomps: # if here it does have sub-components
                        cdict = {} # so store those in __pyomo_components__
                        edict["__pyomo_components__"] = cdict
                    hascomps = True # only make __pyomo_components__ dict once
                    # write each sub-component
                    _write_component(sd=cdict, o=o2, wts=wts, count=count,
                                     lookup=lookup, suffixes=suffixes)

def component_data_to_dict(o, wts):
    """
    Component data to a dict.
    """
    el = o
    alist, ff = wts.get_data_class_attr_list(el) # get attributes
    if alist is None: return # if None then skip writing
    edict = {} # if not writing suffixes don't need ids
    for a in alist: # store desired attributes
        edict[a] = getattr(el, a)
    hascomps = False # Has sub-components (like a Block would have)
    if _may_have_subcomponents(el): # block or block like component
        for o2 in el.component_objects(descend_into=False):
            # loop through sub-components
            if not hascomps: # if here it does have sub-components
                cdict = {} # so store those in __pyomo_components__
                edict["__pyomo_components__"] = cdict
            hascomps = True # only make __pyomo_components__ dict once
            # write each sub-component
            _write_component(sd=cdict, o=o2, wts=wts)
    return edict

def to_json(o, fname=None, human_read=False, wts=None, metadata={}, gz=None,
            return_dict=False, return_json_string=False):
    """
    Save the state of a model to a Python dictionary, and optionally dump it
    to a json file.  To load a model state, a model with the same structure must
    exist.  The model itself cannot be recreated from this.

    Args:
        o: The Pyomo component object to save.  Usually a Pyomo model, but could
            also be a subcomponent of a model (usually a sub-block).
        fname: json file name to save model state, if None only create
            python dict
        gz: If fname is given and gv is True gzip the json file. The default is
            True if the file name ends with '.gz' otherwise False.
        human_read: if True, add indents and spacing to make the json file more
            readable, if false cut out whitespace and make as compact as
            possilbe
        metadata: A dictionary of addtional metadata to add.
        wts: is What To Save, this is a StoreSpec object that specifies what
            object types and attributes to save.  If None, the default is used
            which saves the state of the compelte model state.
        metadata: addtional metadata to save beyond the standard format_version,
            date, and time.
        return_dict: default is False if true returns a dictionary representation
        return_json_string: default is False returns a json string

    Returns:
        If return_dict is True returns a dictionary serialization of the Pyomo
        component.  If return_dict is False and return_json_string is True
        returns a json string dump of the dict.  If fname is given the dictionary
        is also written to a json file.  If gz is True and fname is given, writes
        a gzipped json file.
    """
    if gz is None:
        if isinstance(fname, str):
            gz = fname.endswith(".gz")
        else:
            gz = False

    suffixes = []
    lookup = {}
    count = Counter()
    start_time = time.time()
    if wts is None:
        wts = StoreSpec()
    now = datetime.datetime.now()
    sd={"__metadata__":{
            "format_version":__format_version__,
            "date":datetime.date.isoformat(now.date()),
            "time":datetime.time.isoformat(now.time()),
            "other":metadata}}
    # first write the component
    _write_component(sd, o, wts, count, suffixes=suffixes, lookup=lookup)
    for s in suffixes:
        _write_component_data(**s)
    pdict = {}
    sd["__metadata__"]["__performance__"] = pdict
    pdict["n_components"] = count.count
    dict_time = time.time()
    pdict["etime_make_dict"] = dict_time - start_time
    # This returns the dict but if fname is specified also save to json file
    dump_kw = {'indent': 2} if human_read else {'separators': (',', ':')}
    if fname is not None:
        if gz:
            with gzip.open(fname, 'w') as f:
                f.write(json.dumps(sd, **dump_kw).encode('utf-8'))
        else:
            with open(fname, "w") as f:
                json.dump(sd, f, **dump_kw)
    file_time = time.time()
    # unfortunatly I can't write how long it took to write the file in the file
    pdict["etime_write_file"] = file_time - dict_time
    if return_dict:
        # In interactive environments returning the dict can cuase it to print
        # an extreemly large amount of stuff.  So added this option to make sure
        # it's really what you want.
        return sd
    elif return_json_string:
        return json.dumps(sd, **dump_kw)
    else:
        return None

def _read_component(sd, o, wts, lookup={}, suffixes={}):
    """
    Read a component dictionary into a model
    """
    alist, ff = wts.get_class_attr_list(o)
    if alist is None: return
    oname = o.getname(fully_qualified=False)
    try:
        odict = sd[oname]
    except KeyError as e:
        if wts.ignore_missing:
            return
        else:
            raise(e)
    if ff is not None:
        alist = ff(o, odict)
    if wts.include_suffix:
        lookup[odict['__id__']] = o
    for a in alist:
        try:
            if a in wts.read_cbs:
                if wts.read_cbs[a] is None:
                    pass
                else:
                    wts.read_cbs[a](o, odict[a])
            else:
                setattr(o, a, odict[a])
        except KeyError as e:
            if wts.ignore_missing:
                return
            else:
                raise(e)
    if isinstance(o, Suffix):
        if wts.include_suffix: # make a dict of suffixes to read at the end
            if wts.suffix_filter is None or oname in wts.suffix_filter:
                suffixes[odict['__id__']] = odict["data"] # is populated
    else: # read nonsufix component data
        _read_component_data(odict["data"], o, wts,
                             lookup=lookup, suffixes=suffixes)

def _read_component_data(sd, o, wts, lookup={}, suffixes={}):
    """
    Read a Pyomo component's data in from a dict.

    Args:
        sd: dictionary to read from
        o: Pyomo component whoes data to read
        wts: StoreSpec object specifying what to read in
        lookup: a lookup table for id to componet for reading suffixes
        suffixes: a list of suffixes put off reading until end

    Returns:
        None
    """
    alist = [] # list of attributes to read
    c = 0 # counter of data items in component
    try:
        item_keys = o.keys()
    except AttributeError:
        item_keys = [None]
    for key in item_keys:
        if key is None and isinstance(o, ComponentData) \
            and not isinstance(o, Component):
            el = o
        else:
            el = o[key]
        if c == 0: # if first data item assume all itmes are same and get alist
            alist, ff = wts.get_data_class_attr_list(el) #ff is fileter function
            if alist is None: return #skip reading this type
        c += 1
        try: # get data from dict
            edict = sd[repr(key)]
        except KeyError as e: # data was missing either ignore or raise except
            if wts.ignore_missing:
                return # if ignore missing option its okay
            else:
                raise(e) # else raise exception
        if ff is not None: # if a filer function was given, use it to make a
            # new a list based on the model and whats stored for the state
            # this lets you contionally load things, for example only load
            # values for unfixed variables.
            alist = ff(o, edict)
        if wts.include_suffix: # if loading suffixes make lookup table id to item
            lookup[edict['__id__']] = el
        for a in alist: # read in desired attributes
            try:
                if a in wts.read_cbs:
                    if wts.read_cbs[a] is None:
                        pass
                    else:
                        wts.read_cbs[a](el, edict[a])
                else: # directly set an attribute
                    setattr(el, a, edict[a])
            except KeyError as e: # attribute missing
                if wts.ignore_missing:
                    return # if ignore option then is okay
                else:
                    raise(e) # otherwise raise an exception
        if _may_have_subcomponents(el) and "__pyomo_components__" in edict:
            # read sub-components of block-like
            for o2 in el.component_objects(descend_into=False):
                # recursive read here
                _read_component(edict["__pyomo_components__"], o2, wts,
                                lookup=lookup, suffixes=suffixes)

def component_data_from_dict(sd, o, wts):
    """
    Component data to a dict.
    """
    el = o
    alist = [] # list of attributes to read
    alist, ff = wts.get_data_class_attr_list(el) #ff is fileter function
    if alist is None: return #skip reading this type
    edict = sd
    if ff is not None:
        alist = ff(o, edict)
    for a in alist: # read in desired attributes
        try:
            if a in wts.read_cbs: # use a callback
                wts.read_cbs[a](el, edict[a])
            else: # directly set an attribute
                setattr(el, a, edict[a])
        except KeyError as e: # attribute missing
            if wts.ignore_missing:
                return # if ignore option then is okay
            else:
                raise(e) # otherwise raise an exception
    if _may_have_subcomponents(el): # read sub-components of block-like
        for o2 in el.component_objects(descend_into=False):
            # recursive read here
            _read_component(edict["__pyomo_components__"], o2, wts)

def _read_suffixes(lookup, suffixes):
    """
    Go through the list of suffixes and read the data back in.

    Args:
        lookup: a lookup table to go from id to component
        suffixes: a dictionary with suffix id keys and value dict value
    Returns:
        None
    """
    for uid in suffixes:
        d = suffixes[uid]
        s = lookup[uid] # suffixes keys are ids, so get suffix component
        for key in d: # set values from value dict
            try:
                kc = lookup[int(key)] # use int because json turn keys to string
            except KeyError:
                continue
            s[kc] = d[key]

def from_json(o, sd=None, fname=None, s=None, wts=None, gz=None):
    """
    Load the state of a Pyomo component state from a dictionary, json file, or
    json string.  Must only specify one of sd, fname, or s as a non-None value.
    This works by going through the model and loading the state of each
    sub-compoent of o. If the saved state contains extra information, it is
    ignored.  If the save state doesn't contain an enetry for a model component
    that is to be loaded an error will be raised, unless ignore_missing = True.

    Args:
        o: Pyomo component to for which to load state
        sd: State dictionary to load, if None, check fname and s
        fname: JSON file to load, only used if sd is None
        s: JSON string to load only used if both sd and fname are None
        wts: StoreSpec object specifying what to load
        gz: If True assume the file specified by fname is gzipped. The default is
            True if fname ends with '.gz' otherwise False.

    Returns:
        Dictionary with some perfomance information. The keys are
        "etime_load_file", how long in seconds it took to load the json file
        "etime_read_dict", how long in seconds it took to read models state
        "etime_read_suffixes", how long in seconds it took to read suffixes
    """
    if gz is None:
        if isinstance(fname, str):
            gz = fname.endswith(".gz")
        else:
            gz = False

    # keeping track of elapsed time.  want to make sure I don't do anything
    # that's too slow.
    start_time = time.time()
    # Get the model state dict from one of three sources
    if sd is not None: # Existing Python dict (for in-memory stuff).
        pass
    elif fname is not None: # Read in from a json file
        if gz:
            with gzip.open(fname, 'r') as f:
                fr = f.read()
                sd = json.loads(fr)
        else:
            with open(fname, "r") as f:
                sd = json.load(f) #json file
    elif s is not None: # Use a json string (not really sure if useful)
        sd=json.loads(s) #json string
    else: # Didn't specify at least one source
        raise Exception("Need to specify a data source to load from")
    dict_time = time.time() # To calculate how long it took to read file
    if wts is None: # if no StoreSpec object given use the default, which should
        wts = StoreSpec() # be the typlical save everything important
    lookup = {} # A dict to use for a lookup tables
    suffixes={} # A list of suffixes delayed to end so lookup is complete
    # Read toplevel componet (is recursive)
    _read_component(sd, o, wts, lookup=lookup, suffixes=suffixes)
    read_time = time.time() # to calc time to read model state minus suffixes
    # Now read in the suffixes
    _read_suffixes(lookup, suffixes)
    suffix_time = time.time() # to calculate time to read suffixes
    pdict = {} # return some perfomance information, to make sure not too slow
    pdict["etime_load_file"] = dict_time - start_time
    pdict["etime_read_dict"] = read_time - dict_time
    pdict["etime_read_suffixes"] = suffix_time - read_time
    return pdict



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for flowsheet_model.

Author: Andrew Lee
"""
import pytest
import inspect
from pyomo.environ import ConcreteModel, Constraint, Set, Var
from pyomo.common.config import ConfigBlock
from idaes.core import (declare_process_block_class, ReactionParameterBlock,
                        ReactionBlockBase, ReactionBlockDataBase,
                        PhysicalParameterBlock, StateBlock,
                        StateBlockData)
from idaes.core.util.exceptions import (PropertyPackageError,
                                        PropertyNotSupportedError)


# -----------------------------------------------------------------------------
# Test ParameterBlock
@declare_process_block_class("PropertyParameterBlock")
class _PropertyParameterBlock(PhysicalParameterBlock):
    def build(self):
        super(_PropertyParameterBlock, self).build()

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties({'prop1': {'method': None, 'units': 'm'},
                            'prop3': {'method': False}})
        obj.add_default_units({'time': 's',
                               'length': 'm',
                               'mass': 'g',
                               'amount': 'mol',
                               'temperature': 'K',
                               'energy': 'J',
                               'holdup': 'mol'})


@declare_process_block_class("ReactionParameterTestBlock")
class _ReactionParameterBlock(ReactionParameterBlock):
    def build(self):
        super(ReactionParameterBlock, self).build()


def test_config_block():
    # Test that PhysicalParameterBlock gets module information
    m = ConcreteModel()
    m.r = ReactionParameterTestBlock()

    assert len(m.r.config) == 2
    assert isinstance(m.r.config.default_arguments, ConfigBlock)
    assert len(m.r.config.default_arguments) == 0


def test_ReactionParameter_NotImplementedErrors():
    # Test that class methods return NotImplementedError
    m = ConcreteModel()
    m.r = ReactionParameterTestBlock()

    with pytest.raises(NotImplementedError):
        m.r.get_metadata()


@declare_process_block_class("ReactionParameterBlock2")
class _ReactionParameterBlock2(ReactionParameterBlock):
    def build(self):
        super(ReactionParameterBlock, self).build()

    @classmethod
    def get_required_properties(self):
        return ['prop1']

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties({'rxn1': {'method': None}})
        obj.add_default_units({'time': 'hr',
                               'length': 'm',
                               'mass': 'g',
                               'amount': 'mol',
                               'temperature': 'K',
                               'energy': 'J',
                               'holdup': 'mol'})

def test_validate_state_block_invalid_units():
    # Test validation of associated PropertyParameterBlock
    m = ConcreteModel()
    m.p = PropertyParameterBlock()
    m.r = ReactionParameterBlock2(default={"property_package": m.p})

    with pytest.raises(PropertyPackageError):
        m.r._validate_property_parameter_units()


@declare_process_block_class("ReactionParameterBlock3")
class _ReactionParameterBlock3(ReactionParameterBlock):
    def build(self):
        super(ReactionParameterBlock, self).build()

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties({'rxn1': {'method': None}})
        obj.add_default_units({'time': 's',
                               'length': 'm',
                               'mass': 'g',
                               'amount': 'mol',
                               'temperature': 'K',
                               'energy': 'J',
                               'holdup': 'mol'})
        obj.add_required_properties({'prop2': 'some'})


def test_validate_state_block_unsupported_prop():
    # Test validation of associated PropertyParameterBlock
    m = ConcreteModel()
    m.p = PropertyParameterBlock()
    m.r = ReactionParameterBlock3(default={"property_package": m.p})

    with pytest.raises(PropertyPackageError):
        m.r._validate_property_parameter_properties()


@declare_process_block_class("ReactionParameterBlock4")
class _ReactionParameterBlock4(ReactionParameterBlock):
    def build(self):
        super(ReactionParameterBlock, self).build()

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties({'rxn1': {'method': None}})
        obj.add_default_units({'time': 's',
                               'length': 'm',
                               'mass': 'g',
                               'amount': 'mol',
                               'temperature': 'K',
                               'energy': 'J',
                               'holdup': 'mol'})
        obj.add_required_properties({'prop3': 'some'})


def test_validate_state_block_unsupported_prop_False():
    # Test validation of associated PropertyParameterBlock
    m = ConcreteModel()
    m.p = PropertyParameterBlock()
    m.r = ReactionParameterBlock4(default={"property_package": m.p})

    with pytest.raises(PropertyPackageError):
        m.r._validate_property_parameter_properties()


@declare_process_block_class("ReactionParameterBlock5")
class _ReactionParameterBlock5(ReactionParameterBlock):
    def build(self):
        super(ReactionParameterBlock, self).build()

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties({'rxn1': {'method': None}})
        obj.add_default_units({'time': 's',
                               'length': 'm',
                               'mass': 'g',
                               'amount': 'mol',
                               'temperature': 'K',
                               'energy': 'J',
                               'holdup': 'mol'})
        obj.add_required_properties({'prop1': 'km'})


def test_validate_state_block_req_prop_wrong_units():
    # Test validation of associated PropertyParameterBlock
    m = ConcreteModel()
    m.p = PropertyParameterBlock()
    m.r = ReactionParameterBlock5(default={"property_package": m.p})

    with pytest.raises(PropertyPackageError):
        m.r._validate_property_parameter_properties()


@declare_process_block_class("ReactionParameterBlock6")
class _ReactionParameterBlock6(ReactionParameterBlock):
    def build(self):
        super(ReactionParameterBlock, self).build()

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties({'rxn1': {'method': None}})
        obj.add_default_units({'time': 's',
                               'length': 'm',
                               'mass': 'g',
                               'amount': 'mol',
                               'temperature': 'K',
                               'energy': 'J',
                               'holdup': 'mol'})
        obj.add_required_properties({'prop1': 'm'})


def test_ReactionParameterBase_build():
    # Test that ReactionParameterBlock builds correctly
    m = ConcreteModel()
    m.p = PropertyParameterBlock()
    m.r = ReactionParameterBlock6(default={"property_package": m.p})
    super(_ReactionParameterBlock6, m.r).build()


# -----------------------------------------------------------------------------
# Test ReactionBlockBase
@declare_process_block_class("ReactionBlock",
                             block_class=ReactionBlockBase)
class ReactionBlockData(ReactionBlockDataBase):
    def build(self):
        super(ReactionBlockDataBase, self).build()


def test_ReactionBlockBase_initialize():
    # Test that ReactionBlockBase initialize method raises NotImplementedError
    m = ConcreteModel()
    # Need to index block so that it does not do multiple inheritance
    m.s = Set(initialize=[1, 2])
    m.r = ReactionBlock(m.s)

    with pytest.raises(NotImplementedError):
        m.r.initialize()


def test_ReactionBlockBase_report():
    # Test that ReactionBlockBase report method raises NotImplementedError
    m = ConcreteModel()
    # Need to index block so that it does not do multiple inheritance
    m.s = Set(initialize=[1, 2])
    m.r = ReactionBlock(m.s)

    with pytest.raises(NotImplementedError):
        m.r.report()


# -----------------------------------------------------------------------------
# Test ReactionBlockDataBase
def test_StateBlock_config():
    # Test that ReactionBlockDataBase config has correct arguments
    m = ConcreteModel()
    m.p = ReactionBlock()

    assert len(m.p.config) == 3
    assert hasattr(m.p.config, "parameters")
    assert hasattr(m.p.config, "state_block")
    assert hasattr(m.p.config, "has_equilibrium")

    m.p.config.has_equilibrium = True
    m.p.config.has_equilibrium = False
    with pytest.raises(ValueError):
        m.p.config.has_equilibrium = 'foo'
    with pytest.raises(ValueError):
        m.p.config.has_equilibrium = 10


@declare_process_block_class("TestStateBlock", block_class=StateBlock)
class StateTestBlockData(StateBlockData):
    def build(self):
        super(StateBlockData, self).build()


def test_validate_state_block_fail():
    # Test that ReactionBlockDataBase validate_state_block returns error
    m = ConcreteModel()
    m.p = PropertyParameterBlock()
    m.p2 = PropertyParameterBlock()

    m.pb = TestStateBlock(default={"parameters": m.p2})

    m.r = ReactionParameterBlock6(default={"property_package": m.p})
    super(_ReactionParameterBlock6, m.r).build()

    m.rb = ReactionBlock(default={"parameters": m.r,
                                  "state_block": m.pb})

    with pytest.raises(PropertyPackageError):
        m.rb._validate_state_block()


@declare_process_block_class("ReactionBlock2",
                             block_class=ReactionBlockBase)
class ReactionBlockData2(ReactionBlockDataBase):
    def build(self):
        super(ReactionBlockData2, self).build()


def test_build():
    # Test that ReactionBlockDataBase builds correctly with good argumnets
    m = ConcreteModel()
    m.p = PropertyParameterBlock()

    m.pb = TestStateBlock(default={"parameters": m.p})

    m.r = ReactionParameterBlock6(default={"property_package": m.p})
    super(_ReactionParameterBlock6, m.r).build()

    m.rb = ReactionBlock2(default={"parameters": m.r,
                                   "state_block": m.pb})

    assert hasattr(m.rb, "state_ref")
    assert m.rb._params == m.rb.config.parameters


# -----------------------------------------------------------------------------
# Test reaction __getattr__ method
@declare_process_block_class("Parameters")
class _Parameters(ReactionParameterBlock):
    def build(self):
        super(ReactionParameterBlock, self).build()
        frm = inspect.stack()[1]
        self.__package_module = inspect.getmodule(frm[0])

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties({'a': {'method': 'a_method'},
                            'recursion1': {'method': '_recursion1'},
                            'recursion2': {'method': '_recursion2'},
                            'not_callable': {'method': 'test_obj'},
                            'raise_exception': {'method': '_raise_exception'},
                            'not_supported': {'method': False},
                            'does_not_create_component': {
                                    'method': '_does_not_create_component'}})


@declare_process_block_class("Reaction", block_class=ReactionBlockBase)
class _Reaction(ReactionBlockDataBase):
    def build(self):
        super(ReactionBlockDataBase, self).build()

        self.test_obj = 1

    def a_method(self):
        self.a = Var(initialize=1)

    def _recursion1(self):
        self.recursive_cons1 = Constraint(expr=self.recursion2 == 1)

    def _recursion2(self):
        self.recursive_cons2 = Constraint(expr=self.recursion1 == 1)

    def _raise_exception(self):
        raise Exception()

    def _does_not_create_component(self):
        pass


@pytest.fixture()
def m():
    m = ConcreteModel()
    m.pb = Parameters()
    m.p = Reaction(default={"parameters": m.pb})

    return m


def test_getattr_add_var(m):
    assert isinstance(m.p.a, Var)
    assert m.p.a.value == 1


def test_getattr_protected(m):
    with pytest.raises(PropertyPackageError):
        # Call a protected component that does not exist
        m.p.cons = Constraint(expr=m.p._foo == 1)


def test_getattr_recursion(m):
    with pytest.raises(PropertyPackageError):
        # Call a component that triggers a recursive loop of calls
        m.p.cons = Constraint(expr=m.p.recursion1 == 1)


def test_getattr_does_not_exist(m):
    with pytest.raises(PropertyNotSupportedError):
        m.p.cons = Constraint(expr=m.p.does_not_exist == 1)


def test_getattr_not_callable(m):
    with pytest.raises(PropertyPackageError):
        m.p.cons = Constraint(expr=m.p.not_callable == 1)


def test_getattr_not_supported(m):
    with pytest.raises(PropertyNotSupportedError):
        m.p.cons = Constraint(expr=m.p.not_supported == 1)


def test_getattr_raise_exception(m):
    with pytest.raises(Exception):
        m.p.cons = Constraint(expr=m.p.raise_exception == 1)


# TODO : Need a test for cases where method does not create property
#def test_getattr_does_not_create_component(m):
#    with pytest.raises(PropertyPackageError):
#        m.p.cons = Constraint(expr=m.p.does_not_create_component == 1)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Example property package for the VLE calucations for a Benzene-Toluene-o-Xylene
system. If using the activity coefficient models (NRTL or Wilson), the user is
expected to provide the paramters necessary for these models. Please note that
these parameters are declared as variables here to allow for use in a parameter
estimation problem if the VLE data is available.
"""

# Import Pyomo libraries
from pyomo.environ import Param, NonNegativeReals, Set
from pyomo.common.config import ConfigValue, In

# Import IDAES cores
from idaes.core import declare_process_block_class, PhysicalParameterBlock
from idaes.core.util.misc import extract_data

from idaes.property_models.activity_coeff_models.activity_coeff_prop_pack \
    import ActivityCoeffParameterData
from idaes.logger import getIdaesLogger


# Some more inforation about this module
__author__ = "Andrew Lee, Jaffer Ghouse"
__version__ = "0.0.1"


# Set up logger
_log = getIdaesLogger(__name__)


@declare_process_block_class("MethaneParameterBlock")
class MethaneParameterData(ActivityCoeffParameterData):
    # Methane combstion only considers and ideal vapor phase, so need to
    # overload the user-selection of activity coefficient model and valid
    # phases. Do this by creating our own Config block with limited choices.
    CONFIG = PhysicalParameterBlock.CONFIG()

    CONFIG.declare("activity_coeff_model", ConfigValue(
        default="Ideal",
        domain=In(["Ideal"]),
        description="Methane combustion supports ideal gas only"))

    CONFIG.declare("state_vars", ConfigValue(
        default="FTPz",
        domain=In(["FTPz", "FcTP"]),
        description="Flag indicating the choice for state variables",
        doc="""Flag indicating the choice for state variables to be used
    for the state block, and thus corresponding constraints  should be
    included,
    **default** - FTPz
    **Valid values:** {
    **"FTPx"** - Total flow, Temperature, Pressure and Mole fraction,
    **"FcTP"** - Component flow, Temperature and Pressure}"""))

    CONFIG.declare("valid_phase", ConfigValue(
        default="Vap",
        domain=In(["Vap"]),
        description="Methane combustion supports ideal gas only"))

    def build(self):
        '''
        Callable method for Block construction.
        '''
        super(MethaneParameterData, self).build()

        # Component list - a list of component identifiers
        self.component_list = Set(initialize=['H2', 'N2', 'O2', 'CH4',
                                              'CO', 'CO2', 'H2O', 'NH3'])

        # List of components in each phase (optional)
        self.phase_comp = {"Vap": self.component_list}

        # List of all chemical elements that constitute the chemical species
        self.element_list = Set(initialize=['H', 'N', 'O', 'C'])

        # Elemental composition of all species
        self.element_comp = {'H2': {'H': 2, 'N': 0, 'O': 0, 'C': 0},
                             'N2': {'H': 0, 'N': 2, 'O': 0, 'C': 0},
                             'O2': {'H': 0, 'N': 0, 'O': 2, 'C': 0},
                             'CH4': {'H': 4, 'N': 0, 'O': 0, 'C': 1},
                             'CO': {'H': 0, 'N': 0, 'O': 1, 'C': 1},
                             'CO2': {'H': 0, 'N': 0, 'O': 2, 'C': 1},
                             'H2O': {'H': 2, 'N': 0, 'O': 1, 'C': 0},
                             'NH3': {'H': 3, 'N': 1, 'O': 0, 'C': 0}}

        # Thermodynamic reference state
        self.pressure_reference = Param(mutable=True,
                                        default=101325,
                                        doc='Reference pressure [Pa]')
        self.temperature_reference = Param(mutable=True,
                                           default=1500,
                                           doc='Reference temperature [K]')

        # Gas Constant
        self.gas_const = Param(within=NonNegativeReals,
                               mutable=False,
                               default=8.314,
                               doc='Gas Constant [J/mol.K]')

#        # Source: The Properties of Gases and Liquids (1987)
#        # 4th edition, Chemical Engineering Series - Robert C. Reid
#        mw_comp_data = {'benzene': 78.1136E-3,
#                        'toluene': 92.1405E-3,
#                        'o-xylene': 106.167e-3}
#
#        self.mw_comp = Param(self.component_list,
#                             mutable=False,
#                             initialize=extract_data(mw_comp_data),
#                             doc="molecular weight Kg/mol")

        # Constants for specific heat capacity, enthalpy
        # Sources: The Properties of Gases and Liquids (1987)
        #         4th edition, Chemical Engineering Series - Robert C. Reid
        CpIG_data = {('Vap', 'CH4', 'A'): 1.925e1,
                     ('Vap', 'CH4', 'B'): 5.213e-2,
                     ('Vap', 'CH4', 'C'): 1.197e-5,
                     ('Vap', 'CH4', 'D'): -1.132e-8,
                     ('Vap', 'CH4', 'E'): 0,
                     ('Vap', 'CO', 'A'): 3.087e1,
                     ('Vap', 'CO', 'B'): -1.285e-2,
                     ('Vap', 'CO', 'C'): 2.789e-5,
                     ('Vap', 'CO', 'D'): -1.272e-8,
                     ('Vap', 'CO', 'E'): 0,
                     ('Vap', 'CO2', 'A'): 1.980e1,
                     ('Vap', 'CO2', 'B'): 7.344e-2,
                     ('Vap', 'CO2', 'C'): -5.602e-5,
                     ('Vap', 'CO2', 'D'): 1.715e-8,
                     ('Vap', 'CO2', 'E'): 0,
                     ('Vap', 'H2', 'A'): 2.714e1,
                     ('Vap', 'H2', 'B'): 9.274e-3,
                     ('Vap', 'H2', 'C'): -1.381e-5,
                     ('Vap', 'H2', 'D'): 7.645e-9,
                     ('Vap', 'H2', 'E'): 0,
                     ('Vap', 'H2O', 'A'): 3.224e1,
                     ('Vap', 'H2O', 'B'): 1.924e-3,
                     ('Vap', 'H2O', 'C'): 1.055e-5,
                     ('Vap', 'H2O', 'D'): -3.596e-9,
                     ('Vap', 'H2O', 'E'): 0,
                     ('Vap', 'N2', 'A'): 3.115e1,
                     ('Vap', 'N2', 'B'): -1.357e-2,
                     ('Vap', 'N2', 'C'): 2.680e-5,
                     ('Vap', 'N2', 'D'): -1.168e-8,
                     ('Vap', 'N2', 'E'): 0,
                     ('Vap', 'NH3', 'A'): 2.731e1,
                     ('Vap', 'NH3', 'B'): 2.383e-2,
                     ('Vap', 'NH3', 'C'): 1.707e-5,
                     ('Vap', 'NH3', 'D'): -1.185e-8,
                     ('Vap', 'NH3', 'E'): 0,
                     ('Vap', 'O2', 'A'): 2.811e1,
                     ('Vap', 'O2', 'B'): -3.680e-6,
                     ('Vap', 'O2', 'C'): 1.746e-5,
                     ('Vap', 'O2', 'D'): -1.065e-8,
                     ('Vap', 'O2', 'E'): 0}

        self.CpIG = Param(self.phase_list, self.component_list,
                          ['A', 'B', 'C', 'D', 'E'],
                          mutable=False,
                          initialize=extract_data(CpIG_data),
                          doc="parameters to compute Cp_comp")

        # Source: NIST Webbook, 9th October 2019
        dh_form = {("Vap", "CH4"): -74600,
                   ("Vap", "CO"): -110530,
                   ("Vap", "CO2"): -393520,
                   ("Vap", "H2"): 0,
                   ("Vap", "H2O"): -241830,
                   ("Vap", "N2"): 0,
                   ("Vap", "NH3"): -45900,
                   ("Vap", "O2"): 0}

        self.dh_form = Param(self.phase_list,
                             self.component_list,
                             mutable=False,
                             initialize=extract_data(dh_form),
                             doc="Heats of formation (J/mol)")

        # Source: NIST Webbook, 9th October 2019
        ds_form = {("Vap", "CH4"): 186.25,
                   ("Vap", "CO"): 197.66,
                   ("Vap", "CO2"): 213.79,
                   ("Vap", "H2"): 130.68,
                   ("Vap", "H2O"): 188.84,
                   ("Vap", "N2"): 191.61,
                   ("Vap", "NH3"): 192.77,
                   ("Vap", "O2"): 205.15}

        self.ds_form = Param(self.phase_list,
                             self.component_list,
                             mutable=False,
                             initialize=extract_data(ds_form),
                             doc="Entropies of formation (J/mol.K)")



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for idaes.dmf.resource2 module
"""
# stdlib
import datetime
import json
import logging
import math
import os
import shutil
import sys
import tempfile

# third-party
import pendulum
import pytest

# local
from idaes.dmf import resource, propdata

# for testing
from .util import init_logging

__author__ = "Dan Gunter <dkgunter@lbl.gov>"

if sys.platform.startswith("win"):
    pytest.skip("skipping DMF tests on Windows", allow_module_level=True)

init_logging()
_log = logging.getLogger(__name__)

test_version = "1.2.3"


@pytest.fixture
def example_resource():
    r = resource.Resource()
    r.v["version_info"]["version"] = test_version
    r.v["collaborators"] = [
        {"name": "Clark Kent", "email": "ckent@dailyplanet.com"},
        {"name": "Superman", "email": "sman@fortress.solitude.org"},
    ]
    r.v["sources"].append(
        {
            "isbn": "978-0201616224",
            "source": 'Hunt, A. and Thomas, D., "The Pragmatic Programmer", '
            "Addison-Wesley, 1999",
            "date": "1999-01-01",
        }
    )
    r.v["codes"].append(
        {
            "type": "function",
            "name": "test_resource_full",
            "desc": "The test function",
            "location": __file__,
            "version": test_version,
        }
    )
    r.v["datafiles"].append({"path": "/etc/passwd"})
    r.v["aliases"] = ["test_resource_full"]
    r.v["tags"] = ["test", "resource"]
    # r.relations  -- deal with this separately
    r.data = {"arbitrary": {"values": [1, 2, 3]}}
    return r


@pytest.fixture
def default_resource():
    return resource.Resource()


def test_resource_roundtrip(example_resource):
    """Build up a resource with all attributes,
    then make sure it serializes and deserializes.
    """
    r = example_resource  # alias
    # make sure we can serialize it
    r_json = json.dumps(r.v, indent=2)
    # make sure we can deserialize its JSON
    r2_value = json.loads(r_json)
    # reconstruct from the deserialized JSON
    r2 = resource.Resource(value=r2_value)
    # compare them
    assert r2.v["version_info"] == r.v["version_info"]
    assert r2.v["sources"][0]["date"] == r.v["sources"][0]["date"]
    assert r2.data["arbitrary"]["values"][0] == 1


_propd = [
    {
        "name": "Viscosity Value",
        "units": "mPa-s",
        "values": [2.6, 6.2],
        "error_type": "absolute",
        "errors": [0.06, 0.004],
        "type": "property",
    },
    {"name": "r", "units": "", "values": [0.2, 1000], "type": "state"},
]

_propm = {
    "datatype": "MEA",
    "info": "J. Chem. Eng. Data, 2009, Vol 54, pg. 3096-30100",
    "notes": "r is MEA weight fraction in aqueous soln.",
    "authors": "Amundsen, T.G., Lars, E.O., Eimer, D.A.",
    "title": "Density and Viscosity of Monoethanolamine + .etc.",
    "date": "1970-01-01",
}


def test_property_data_resource():
    r = resource.Resource()
    r.data = {"data": _propd, "meta": _propm}
    tbl = propdata.PropertyTable(data=r.data["data"], metadata=r.data["meta"])
    assert tbl.data.properties[0]
    assert tbl.data.properties[0]["values"] == _propd[0]["values"]
    assert tbl.data.properties[0]["errors"] == _propd[0]["errors"]
    assert tbl.metadata[0].title == _propm["title"]


def test_validate_default(default_resource):
    default_resource.validate()


def test_validate_example(example_resource):
    example_resource.validate()


def test_validate_preprocess(default_resource):
    r = default_resource
    r.validate()
    r.v["created"] = "2012-01-01"
    r.v["modified"] = "2012-01-01"
    r.v["version_info"]["created"] = "2011-01-01"
    r.validate()
    with pytest.raises(ValueError):
        r.v["created"] = "None of your business"
        r.validate()


@pytest.fixture
def tmpd():
    d = tempfile.mkdtemp(prefix="test_resource_", suffix=".idaes")
    yield d
    shutil.rmtree(d)


def test_get_datafiles_relative(default_resource, tmpd):
    r = default_resource
    r.v["datafiles_dir"] = tmpd  # paths are now relative to this
    paths = set()
    for i in range(3):
        name = "file{:d}".format(i)
        path = os.path.join(tmpd, name)
        open(path, "w").write("hello {:d}\n".format(i))
        r.v["datafiles"].append({"path": name})
        paths.add(path)
    # check that these, and only these, files are returned
    for f in r.get_datafiles():
        assert f.name in paths
        paths.remove(f.name)
    assert len(paths) == 0  # all were returned


def test_get_datafiles_absolute(default_resource, tmpd):
    r = default_resource
    paths = set()
    for i in range(3):
        name = "file{:d}".format(i)
        path = os.path.join(tmpd, name)
        open(path, "w").write("hello {:d}\n".format(i))
        # note: unlike "..._relative()", add the full path
        r.v["datafiles"].append({"path": path})
        paths.add(path)
    # check that these, and only these, files are returned
    for f in r.get_datafiles():
        assert f.name in paths
        paths.remove(f.name)
    assert len(paths) == 0  # all were returned


def test_create_relation(default_resource, example_resource):
    r1, r2 = default_resource, example_resource
    relation = resource.Triple(r1, resource.PR_USES, r2)
    resource.create_relation(relation)
    with pytest.raises(ValueError):  # duplicate will raise ValueError
        resource.create_relation(relation)
    assert len(r1.v["relations"]) == 1
    assert len(r2.v["relations"]) == 1
    assert r1.v["relations"][0]["role"] == "subject"
    assert r2.v["relations"][0]["role"] == "object"
    assert r1.v["relations"][0]["identifier"] == r2.v[r2.ID_FIELD]
    assert r2.v["relations"][0]["identifier"] == r1.v[r2.ID_FIELD]
    # some errors
    with pytest.raises(ValueError):
        resource.create_relation_args("foo", "bad predicate", "bar")
    # delete relation from subject to test duplicate check for object
    r1.v["relations"] = []
    with pytest.raises(ValueError):  # dup raises ValueError
        resource.create_relation(relation)


def test_repr(example_resource):
    txt = example_resource._repr_text_()
    assert len(txt) > 0


def test_date_float():
    now = pendulum.now()
    now_float = now.timestamp()
    assert resource.date_float(now) == now_float
    # tuples
    now_tuple = tuple(list(now.timetuple()[:6]) + [now.microsecond, now.tzinfo])
    assert resource.date_float(now_tuple) == now_float
    with pytest.raises(ValueError):
        resource.date_float(tuple("garbage"))
    # datetime
    assert resource.date_float(datetime.datetime(*now_tuple)) == now_float
    # strings
    with pytest.raises(ValueError):
        resource.date_float("garbage")
    assert resource.date_float(now.isoformat()) == now_float
    # int/float
    assert resource.date_float(now_float) == now_float
    assert resource.date_float(int(now_float)) == math.floor(now_float)
    #    with pytest.raises(ValueError):
    #        resource.date_float(1e12)
    #    with pytest.raises(ValueError):
    #        resource.date_float(1000000000000)
    # none
    with pytest.raises(ValueError):
        resource.date_float(None)


def test_version_list():
    f = resource.version_list
    # any tuple prefix is fine
    assert f(1) == f((1,)) == f((1, 0)) == f((1, 0, 0)) == f((1, 0, 0, ""))
    # same thing with a string is fine
    assert f("1") == f("1.0") == f("1.0.0")
    # check all bad possibilities
    with pytest.raises(ValueError):
        f(())
    with pytest.raises(ValueError):
        f(1.0)
    with pytest.raises(ValueError):
        f(("a",))
    with pytest.raises(ValueError):
        f(("a", "b"))
    with pytest.raises(ValueError):
        f((1, 2, 3, None))
    with pytest.raises(ValueError):
        f((1, 2, 3, datetime.datetime))
    with pytest.raises(ValueError):
        f("1.2.3.4")  # last bit can't start with '.'
    assert f("1.2.3RC3") == [1, 2, 3, "RC3"]  # extra
    assert f("1.2.3-RC3") == [1, 2, 3, "RC3"]  # stripped "-"


def test_format_version():
    assert resource.format_version([1, 2, 3, "RC3"]) == "1.2.3-RC3"


def test_identifier_str():
    assert len(resource.identifier_str()) > 1
    assert resource.identifier_str("0" * 32) == "0" * 32
    with pytest.raises(ValueError):
        resource.identifier_str("foobar")
    with pytest.raises(ValueError):
        resource.identifier_str("0" * 31 + "X")


def test_dirty_bit():
    dd = resource.Dict({"value": 1})
    assert dd.is_dirty()
    dd.set_clean()
    assert not dd.is_dirty()
    dd["value"] = 2
    assert dd.is_dirty()
    dd.set_clean()
    assert not dd.is_dirty()


def test_validate_onlywhendirty(default_resource):
    r = default_resource
    assert r._validations == 0
    r.validate()
    assert r._validations == 1
    r.validate()
    assert r._validations == 1
    r.v["tags"] = ["one tag"]
    r.validate()
    assert r._validations == 2
    r.validate()
    assert r._validations == 2


def test_triple_from_resource_relations():
    i = "cookie monster"
    j = "cookies"
    # cookie monster likes cookies
    d = {
        resource.RR_ROLE: resource.RR_SUBJ,
        resource.RR_ID: j,
        resource.RR_PRED: "likes",
    }
    t = resource.triple_from_resource_relations(i, d)
    assert t.subject == i
    assert t.predicate == "likes"
    assert t.object == j
    # cookies are liked by cookie monster
    # result is same relation triple as above
    d = {
        resource.RR_ROLE: resource.RR_OBJ,
        resource.RR_ID: i,
        resource.RR_PRED: "likes",
    }
    t = resource.triple_from_resource_relations(j, d)
    assert t.subject == i
    assert t.predicate == "likes"
    assert t.object == j




##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
The process_block module simplifies inheritance of Pyomo blocks. The main
reason to subclass a Pyomo block is to create a block that comes with
pre-defined model equations. This is used in the IDAES modeling framework to
create modular process model blocks.
"""

import sys
import logging

from pyomo.common.config import ConfigBlock
from pyomo.environ import Block

__author__ = "John Eslick"
__all__ = ['ProcessBlock', 'declare_process_block_class']


def _rule_default(b, *args):
    """
    Default rule for ProcessBlock, which calls build(). A different rule can
    be specified to add additional build steps, or to not call build at all
    using the normal rule argument to ProcessBlock init.
    """
    try:
        b.build()
    except Exception as e:
        logging.getLogger(__name__).exception(
            "Failure in build: {}".format(b))
        raise e

_process_block_docstring = """
    Args:
        rule (function): A rule function or None. Default rule calls build().
        concrete (bool): If True, make this a toplevel model. **Default** - False.
        ctype (str): Pyomo ctype of the block.  **Default** - "Block"
        default (dict): Default ProcessBlockData config{}
        initialize (dict): ProcessBlockData config for individual elements. Keys
            are BlockData indexes and values are dictionaries described under the
            "default" argument above.
        idx_map (function): Function to take the index of a BlockData element and
            return the index in the initialize dict from which to read arguments.
            This can be provided to overide the default behavior of matching the
            BlockData index exactly to the index in initialize.
    Returns:
        ({}) New instance
    """

_config_block_keys_docstring = """

            ..

            Keys
{}
            ..
"""

def _process_kwargs(o, kwargs):
    kwargs.setdefault("rule", _rule_default)
    o._block_data_config_default = kwargs.pop("default", None)
    o._block_data_config_initialize = ConfigBlock(implicit=True)
    o._block_data_config_initialize.set_value(kwargs.pop("initialize", None))
    o._idx_map = kwargs.pop("idx_map", None)


class _IndexedProcessBlockMeta(type):
    """Metaclass used to create an indexed model class."""

    def __new__(meta, name, bases, dct):
        def __init__(self, *args, **kwargs):
            _process_kwargs(self, kwargs)
            bases[0].__init__(self, *args, **kwargs)
        dct["__init__"] = __init__
        dct["__process_block__"] = "indexed"
        return type.__new__(meta, name, bases, dct)


class _ScalarProcessBlockMeta(type):
    """Metaclass used to create a scalar model class."""

    def __new__(meta, name, bases, dct):
        def __init__(self, *args, **kwargs):
            _process_kwargs(self, kwargs)
            bases[0].__init__(self, component=self)
            bases[1].__init__(self, *args, **kwargs)
        dct["__init__"] = __init__
        dct["__process_block__"] = "scalar"
        return type.__new__(meta, name, bases, dct)


class ProcessBlock(Block):
    __doc__ = """
    ProcessBlock is a Pyomo Block that is part of a system to make Pyomo
    Block easier to subclass. The main difference between a Pyomo Block and
    ProcessBlock from the user perspective is that a ProcessBlock has a rule
    assigned by default that calls the build() method for the contained
    ProcessBlockData objects. The default rule can be overridden, but the new
    rule should always call build() for the ProcessBlockData object.
    """ + _process_block_docstring.format("", "ProcessBlock")

    def __new__(cls, *args, **kwds):
        """Create a new indexed or scalar ProcessBlock subclass instance
        depending on whether there are args.  If there are args those should be
        an indexing set."""
        if hasattr(cls, "__process_block__"):
            # __process_block__ is a class attribute created when making an
            # indexed or scalar subclass of ProcessBlock (or subclass thereof).
            # If cls dosen't have it, the indexed or scalar class has not been
            # created yet.
            #
            # You get here after creating a new indexed or scalar class in the
            # next if below. The first time in, cls is a ProcessBlock subclass
            # that is neither indexed or scalar so you go to the if below and
            # create an index or scalar subclass of cls.
            return super(Block, cls).__new__(cls)
        if args == (): # no args so make scalar class
            bname = "_Scalar{}".format(cls.__name__)
            n = _ScalarProcessBlockMeta(bname, (cls._ComponentDataClass, cls),{})
            return n.__new__(n) #calls this __new__() again with scalar class
        else: # args so make indexed class
            bname = "_Indexed{}".format(cls.__name__)
            n = _IndexedProcessBlockMeta(bname, (cls,), {})
            return n.__new__(n) #calls this __new__() again with indexed class

    @classmethod
    def base_class_name(cls):
        """Name given by the user to the ProcessBase class.

        Return:
           (str) Name of the class.
        Raises:
           AttributeError, if no base class name was set, e.g. this class
               was *not* wrapped by the `declare_process_block_class`
               decorator.

        """
        return cls._orig_name

    @classmethod
    def base_class_module(cls):
        """Return module of the associated ProcessBase class.

        Return:
           (str) Module of the class.
        Raises:
           AttributeError, if no base class module was set, e.g. this class
              was *not* wrapped by the `declare_process_block_class` decorator.

        """
        return cls._orig_module


def declare_process_block_class(name, block_class=ProcessBlock, doc=""):
    """
    Declare a new ProcessBlock subclass.

    This is a decorator function for a class definition, where the class is
    derived from Pyomo's _BlockData. It creates a ProcessBlock subclass to
    contain the decorated class. The only requirment is that the subclass of
    _BlockData contain a build() method. The purpose of this decorator is to
    simplify subclassing Pyomo's block class.

    Args:
        name: name of class to create
        block_class: ProcessBlock or a subclass of ProcessBlock, this allows
            you to use a subclass of ProcessBlock if needed. The typical use
            case for Subclassing ProcessBlock is to impliment methods that
            operate on elements of an indexed block.
        doc: Documentation for the class. This should play nice with sphinx.
    Returns:
        Decorator function

    """
    def proc_dec(cls):  # Decorator function
        # create a new class called name from block_class
        try:
            cb_doc = cls.CONFIG.generate_documentation(
                block_start="", block_end="", item_start="%s\n",
                indent_spacing=4, item_body="%s", item_end="\n", width=66)
            cb_doc += "\n"
            cb_doc = '\n'.join(' '*12 + x for x in cb_doc.splitlines())
        except:
            cb_doc = ""
        if cb_doc != "":
            cb_doc = _config_block_keys_docstring.format(cb_doc)
        ds = "\n".join([doc, _process_block_docstring.format(cb_doc, name)])
        c = type(name, (block_class,),
                {"__module__": cls.__module__,
                 "_ComponentDataClass": cls,
                 "__doc__":ds})
        setattr(sys.modules[cls.__module__], name, c)
        setattr(cls, '_orig_name', name)
        setattr(cls, '_orig_module', cls.__module__)
        return cls
    return proc_dec  # return decorator function



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################


def mapminmax(x, *args):
    import numpy as np
    if args == ():
        xmax = np.amax(x, 0)
        xmin = np.amin(x, 0)
    else:
        xmax = args[0]
        xmin = args[1]
    try:
        ndata, ninputs = np.shape(x)
        y = np.ones([ndata, ninputs])
    except Exception:
        ndata = np.size(x)
        ninputs = 1
        y = np.ones([ndata])
    if ninputs > 1:
        # check for constant column
        for j in range(ninputs):
            if (xmin[j] != xmax[j]):
                y[:, j] = 2 * (x[:, j] - xmin[j]) / float(xmax[j] - xmin[j]) - 1
            else:
                y[:, j] = 1.0
    else:
        if (xmin != xmax):
            y = 2 * (x - xmin) / float(xmax - xmin) - 1
        else:
            y[:] = 1.0
    return y, xmax, xmin
